import inspect
import sys
from typing import Annotated, Any, Literal

from pydantic import BaseModel, Field
from typing_extensions import TypeAliasType

CustomConst = Any  # TODO


class ExtensionVal(BaseModel):
    """An extension constant value, that can check it is of a given [CustomType]."""

    v: Literal["Extension"] = "Extension"
    c: tuple[CustomConst]


class FunctionVal(BaseModel):
    """A higher-order function value."""

    v: Literal["Function"] = "Function"
    hugr: Any  # TODO


class Tuple(BaseModel):
    """A tuple."""

    v: Literal["Tuple"] = "Tuple"
    vs: list["Value"]


class Sum(BaseModel):
    """A Sum variant

    For any Sum type where this value meets the type of the variant indicated by the tag
    """

    v: Literal["Sum"] = "Sum"
    tag: int
    value: "Value"


Value = TypeAliasType(
    "Value",
    Annotated[
        ExtensionVal | FunctionVal | Tuple | Sum,
        Field(discriminator="v"),
    ],
)


# Now that all classes are defined, we need to update the ForwardRefs in all type
# annotations. We use some inspect magic to find all classes defined in this file.
classes = inspect.getmembers(
    sys.modules[__name__],
    lambda member: inspect.isclass(member) and member.__module__ == __name__,
)
for _, c in classes:
    if issubclass(c, BaseModel):
        c.model_rebuild()



import ast
from dataclasses import dataclass

from guppylang.ast_util import AstNode, has_empty_body, with_loc
from guppylang.checker.core import Context, Globals
from guppylang.checker.expr_checker import check_call, synthesize_call
from guppylang.checker.func_checker import check_signature
from guppylang.compiler.core import CompiledFunction, CompiledGlobals, DFContainer
from guppylang.error import GuppyError
from guppylang.gtypes import GuppyType, Inst, Subst, type_to_row
from guppylang.hugr.hugr import Hugr, Node, OutPortV, VNode
from guppylang.nodes import GlobalCall


@dataclass
class DeclaredFunction(CompiledFunction):
    """A user-declared function that compiles to a Hugr function declaration."""

    node: VNode | None = None

    @staticmethod
    def from_ast(
        func_def: ast.FunctionDef, name: str, globals: Globals
    ) -> "DeclaredFunction":
        ty = check_signature(func_def, globals)
        if not has_empty_body(func_def):
            raise GuppyError(
                "Body of function declaration must be empty", func_def.body[0]
            )
        return DeclaredFunction(name, ty, func_def, None)

    def check_call(
        self, args: list[ast.expr], ty: GuppyType, node: AstNode, ctx: Context
    ) -> tuple[ast.expr, Subst]:
        # Use default implementation from the expression checker
        args, subst, inst = check_call(self.ty, args, ty, node, ctx)
        return with_loc(node, GlobalCall(func=self, args=args, type_args=inst)), subst

    def synthesize_call(
        self, args: list[ast.expr], node: AstNode, ctx: Context
    ) -> tuple[GlobalCall, GuppyType]:
        # Use default implementation from the expression checker
        args, ty, inst = synthesize_call(self.ty, args, node, ctx)
        return with_loc(node, GlobalCall(func=self, args=args, type_args=inst)), ty

    def add_to_graph(self, graph: Hugr, parent: Node) -> None:
        self.node = graph.add_declare(self.ty, parent, self.name)

    def load(
        self, dfg: DFContainer, graph: Hugr, globals: CompiledGlobals, node: AstNode
    ) -> OutPortV:
        assert self.node is not None
        return graph.add_load_constant(self.node.out_port(0), dfg.node).out_port(0)

    def compile_call(
        self,
        args: list[OutPortV],
        type_args: Inst,
        dfg: DFContainer,
        graph: Hugr,
        globals: CompiledGlobals,
        node: AstNode,
    ) -> list[OutPortV]:
        assert self.node is not None
        # TODO: Hugr should probably allow us to pass type args to `Call`, so we can
        #   avoid loading the function to manually add a `TypeApply`
        if type_args:
            func = graph.add_load_constant(self.node.out_port(0), dfg.node)
            func = graph.add_type_apply(func.out_port(0), type_args, dfg.node)
            call = graph.add_indirect_call(func.out_port(0), args, dfg.node)
        else:
            call = graph.add_call(self.node.out_port(0), args, dfg.node)
        return [call.out_port(i) for i in range(len(type_to_row(self.ty.returns)))]



import ast
import inspect
import sys
import textwrap
from collections.abc import Callable
from types import ModuleType
from typing import Any, Union

from guppylang.ast_util import AstNode, annotate_location
from guppylang.checker.core import Globals, PyScope, TypeVarDecl, qualified_name
from guppylang.checker.func_checker import DefinedFunction, check_global_func_def
from guppylang.compiler.core import CompiledGlobals
from guppylang.compiler.func_compiler import (
    CompiledFunctionDef,
    compile_global_func_def,
)
from guppylang.custom import CustomFunction
from guppylang.declared import DeclaredFunction
from guppylang.error import GuppyError, pretty_errors
from guppylang.gtypes import GuppyType
from guppylang.hugr.hugr import Hugr

PyFunc = Callable[..., Any]
PyFuncDefOrDecl = tuple[bool, PyFunc]


class GuppyModule:
    """A Guppy module that may contain function and type definitions."""

    name: str

    # Whether the module has already been compiled
    _compiled: bool

    # Globals from imported modules
    _imported_globals: Globals
    _imported_compiled_globals: CompiledGlobals

    # Globals for functions and types defined in this module. Only gets populated during
    # compilation
    _globals: Globals
    _compiled_globals: CompiledGlobals

    # Mappings of functions defined in this module
    _func_defs: dict[str, tuple[ast.FunctionDef, PyScope]]
    _func_decls: dict[str, ast.FunctionDef]
    _custom_funcs: dict[str, CustomFunction]

    # When `_instance_buffer` is not `None`, then all registered functions will be
    # buffered in this list. They only get properly registered, once
    # `_register_buffered_instance_funcs` is called. This way, we can associate
    _instance_func_buffer: dict[str, PyFuncDefOrDecl | CustomFunction] | None

    def __init__(self, name: str, import_builtins: bool = True):
        self.name = name
        self._globals = Globals({}, {}, {}, {})
        self._compiled_globals = {}
        self._imported_globals = Globals.default()
        self._imported_compiled_globals = {}
        self._func_defs = {}
        self._func_decls = {}
        self._custom_funcs = {}
        self._compiled = False
        self._instance_func_buffer = None

        # Import builtin module
        if import_builtins:
            import guppylang.prelude.builtins as builtins

            self.load(builtins)

    def load(self, m: Union[ModuleType, "GuppyModule"]) -> None:
        """Imports another Guppy module."""
        self._check_not_yet_compiled()
        if isinstance(m, GuppyModule):
            # Compile module if it isn't compiled yet
            if not m.compiled:
                m.compile()

            # For now, we can only import custom functions
            if any(
                not isinstance(v, CustomFunction) for v in m._compiled_globals.values()
            ):
                raise GuppyError(
                    "Importing modules with defined functions is not supported yet"
                )

            self._imported_globals |= m._globals
            self._imported_compiled_globals |= m._compiled_globals
        else:
            for val in m.__dict__.values():
                if isinstance(val, GuppyModule):
                    self.load(val)

    def register_func_def(
        self, f: PyFunc, instance: type[GuppyType] | None = None
    ) -> None:
        """Registers a Python function definition as belonging to this Guppy module."""
        self._check_not_yet_compiled()
        func_ast = parse_py_func(f)
        if self._instance_func_buffer is not None:
            self._instance_func_buffer[func_ast.name] = (True, f)
        else:
            name = (
                qualified_name(instance, func_ast.name) if instance else func_ast.name
            )
            self._check_name_available(name, func_ast)
            self._func_defs[name] = func_ast, get_py_scope(f)

    def register_func_decl(
        self, f: PyFunc, instance: type[GuppyType] | None = None
    ) -> None:
        """Registers a Python function declaration as belonging to this Guppy module."""
        self._check_not_yet_compiled()
        func_ast = parse_py_func(f)
        if self._instance_func_buffer is not None:
            self._instance_func_buffer[func_ast.name] = (False, f)
        else:
            name = (
                qualified_name(instance, func_ast.name) if instance else func_ast.name
            )
            self._check_name_available(name, func_ast)
            self._func_decls[name] = func_ast

    def register_custom_func(
        self, func: CustomFunction, instance: type[GuppyType] | None = None
    ) -> None:
        """Registers a custom function as belonging to this Guppy module."""
        self._check_not_yet_compiled()
        if self._instance_func_buffer is not None:
            self._instance_func_buffer[func.name] = func
        else:
            if instance:
                func.name = qualified_name(instance, func.name)
            self._check_name_available(func.name, func.defined_at)
            self._custom_funcs[func.name] = func

    def register_type(self, name: str, ty: type[GuppyType]) -> None:
        """Registers an existing Guppy type as belonging to this Guppy module."""
        self._check_not_yet_compiled()
        self._check_type_name_available(name, None)
        self._globals.types[name] = ty

    def register_type_var(self, name: str, linear: bool) -> None:
        """Registers a new type variable"""
        self._check_not_yet_compiled()
        self._check_type_name_available(name, None)
        self._globals.type_vars[name] = TypeVarDecl(name, linear)

    def _register_buffered_instance_funcs(self, instance: type[GuppyType]) -> None:
        assert self._instance_func_buffer is not None
        buffer = self._instance_func_buffer
        self._instance_func_buffer = None
        for f in buffer.values():
            if isinstance(f, CustomFunction):
                self.register_custom_func(f, instance)
            else:
                is_def, pyfunc = f
                if is_def:
                    self.register_func_def(pyfunc, instance)
                else:
                    self.register_func_decl(pyfunc, instance)

    @property
    def compiled(self) -> bool:
        return self._compiled

    @pretty_errors
    def compile(self) -> Hugr | None:
        """Compiles the module and returns the final Hugr."""
        if self.compiled:
            raise GuppyError("Module has already been compiled")

        # Prepare globals for type checking
        for func in self._custom_funcs.values():
            func.check_type(self._imported_globals | self._globals)
        defined_funcs = {
            x: DefinedFunction.from_ast(f, x, self._imported_globals | self._globals)
            for x, (f, _) in self._func_defs.items()
        }
        declared_funcs = {
            x: DeclaredFunction.from_ast(f, x, self._imported_globals | self._globals)
            for x, f in self._func_decls.items()
        }
        self._globals.values.update(self._custom_funcs)
        self._globals.values.update(declared_funcs)
        self._globals.values.update(defined_funcs)

        # Type check function definitions
        checked = {
            x: check_global_func_def(
                f,
                self._imported_globals
                | self._globals
                | Globals({}, {}, {}, self._func_defs[x][1]),
            )
            for x, f in defined_funcs.items()
        }

        # Add declared functions to the graph
        graph = Hugr(self.name)
        module_node = graph.set_root_name(self.name)
        for f in declared_funcs.values():
            f.add_to_graph(graph, module_node)

        # Prepare `FunctionDef` nodes for all function definitions
        def_nodes = {x: graph.add_def(f.ty, module_node, x) for x, f in checked.items()}
        self._compiled_globals |= (
            self._custom_funcs
            | declared_funcs
            | {
                x: CompiledFunctionDef(x, f.ty, f.defined_at, None, def_nodes[x])
                for x, f in checked.items()
            }
        )

        # Compile function definitions to Hugr
        for x, f in checked.items():
            compile_global_func_def(
                f,
                def_nodes[x],
                graph,
                self._imported_compiled_globals | self._compiled_globals,
            )

        self._compiled = True
        return graph

    def contains_function(self, name: str) -> bool:
        """Returns 'True' if the module contains a function with the given name."""
        return name in self._func_defs or name in self._custom_funcs

    def contains_type(self, name: str) -> bool:
        """Returns 'True' if the module contains a type with the given name."""
        return name in self._globals.types or name in self._globals.type_vars

    def _check_not_yet_compiled(self) -> None:
        if self._compiled:
            raise GuppyError(f"The module `{self.name}` has already been compiled")

    def _check_name_available(self, name: str, node: AstNode | None) -> None:
        if self.contains_function(name):
            raise GuppyError(
                f"Module `{self.name}` already contains a function named `{name}`",
                node,
            )

    def _check_type_name_available(self, name: str, node: AstNode | None) -> None:
        if name in self._globals.types:
            raise GuppyError(
                f"Module `{self.name}` already contains a type `{name}`",
                node,
            )

        if name in self._globals.type_vars:
            raise GuppyError(
                f"Module `{self.name}` already contains a type variable `{name}`",
                node,
            )


def parse_py_func(f: PyFunc) -> ast.FunctionDef:
    source_lines, line_offset = inspect.getsourcelines(f)
    source = "".join(source_lines)  # Lines already have trailing \n's
    source = textwrap.dedent(source)
    func_ast = ast.parse(source).body[0]
    file = inspect.getsourcefile(f)
    if file is None:
        raise GuppyError("Couldn't determine source file for function")
    annotate_location(func_ast, source, file, line_offset)
    if not isinstance(func_ast, ast.FunctionDef):
        raise GuppyError("Expected a function definition", func_ast)
    return func_ast


def get_py_scope(f: PyFunc) -> PyScope:
    """Returns a mapping of all variables captured by a Python function.

    Note that this function only works in CPython. On other platforms, an empty
    dictionary is returned.

    Relies on inspecting the `__globals__` and `__closure__` attributes of the function.
    See https://docs.python.org/3/reference/datamodel.html#special-read-only-attributes
    """
    if sys.implementation.name != "cpython":
        return {}

    if inspect.ismethod(f):
        f = f.__func__
    code = f.__code__

    nonlocals: PyScope = {}
    if f.__closure__ is not None:
        for var, cell in zip(code.co_freevars, f.__closure__):
            try:
                value = cell.cell_contents
            except ValueError:
                # The call to `cell_contents` will fail if `var` is a recursive
                # reference to the decorated function
                continue
            nonlocals[var] = value

    return nonlocals | f.__globals__.copy()



import inspect
import sys
from enum import Enum
from typing import Annotated, Any, Literal

from pydantic import (
    BaseModel,
    Field,
    ValidationError,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    WrapValidator,
)
from pydantic_core import PydanticCustomError
from typing_extensions import TypeAliasType


def _json_custom_error_validator(
    value: Any, handler: ValidatorFunctionWrapHandler, _info: ValidationInfo
) -> Any:
    """Simplify the error message to avoid a gross error stemming
    from exhaustive checking of all union options.

    As suggested at
    https://docs.pydantic.dev/latest/concepts/types/#named-recursive-types


    Used to define named recursive alias types.
    """
    try:
        return handler(value)
    except ValidationError as err:
        raise PydanticCustomError(
            "invalid_json",
            "Input is not valid json",
        ) from err


ExtensionId = str
ExtensionSet = list[  # TODO: Set not supported by MessagePack. Is list correct here?
    ExtensionId
]


# --------------------------------------------
# --------------- TypeParam ------------------
# --------------------------------------------


class TypeTypeParam(BaseModel):
    tp: Literal["Type"] = "Type"
    b: "TypeBound"


class BoundedNatParam(BaseModel):
    tp: Literal["BoundedNat"] = "BoundedNat"
    bound: int | None


class OpaqueParam(BaseModel):
    tp: Literal["Opaque"] = "Opaque"
    ty: "Opaque"


class ListParam(BaseModel):
    tp: Literal["List"] = "List"
    param: "TypeParam"


class TupleParam(BaseModel):
    tp: Literal["Tuple"] = "Tuple"
    params: list["TypeParam"]


TypeParam = TypeAliasType(
    "TypeParam",
    Annotated[
        TypeTypeParam | BoundedNatParam | OpaqueParam | ListParam | TupleParam,
        Field(discriminator="tp"),
        WrapValidator(_json_custom_error_validator),
    ],
)


# ------------------------------------------
# --------------- TypeArg ------------------
# ------------------------------------------


class CustomTypeArg(BaseModel):
    typ: None  # TODO
    value: str


class TypeTypeArg(BaseModel):
    tya: Literal["Type"] = "Type"
    ty: "Type"


class BoundedNatArg(BaseModel):
    tya: Literal["BoundedNat"] = "BoundedNat"
    n: int


class OpaqueArg(BaseModel):
    tya: Literal["Opaque"] = "Opaque"
    arg: CustomTypeArg


class SequenceArg(BaseModel):
    tya: Literal["Sequence"] = "Sequence"
    args: list["TypeArg"]


class ExtensionsArg(BaseModel):
    tya: Literal["Extensions"] = "Extensions"
    es: ExtensionSet


TypeArg = TypeAliasType(
    "TypeArg",
    Annotated[
        TypeTypeArg | BoundedNatArg | OpaqueArg | SequenceArg | ExtensionsArg,
        Field(discriminator="tya"),
        WrapValidator(_json_custom_error_validator),
    ],
)


# --------------------------------------------
# --------------- Container ------------------
# --------------------------------------------


class MultiContainer(BaseModel):
    ty: "Type"


class Array(MultiContainer):
    """Known size array of"""

    t: Literal["Array"] = "Array"
    len: int


class TupleType(BaseModel):
    """Product type, known-size tuple over elements of type row"""

    t: Literal["Tuple"] = "Tuple"
    inner: "TypeRow"


class UnitSum(BaseModel):
    """Simple predicate where all variants are empty tuples"""

    t: Literal["Sum"] = "Sum"

    s: Literal["Unit"] = "Unit"
    size: int


class GeneralSum(BaseModel):
    """General sum type that explicitly stores the types of the variants"""

    t: Literal["Sum"] = "Sum"

    s: Literal["General"] = "General"
    row: "TypeRow"


Sum = Annotated[UnitSum | GeneralSum, Field(discriminator="s")]
# ----------------------------------------------
# --------------- ClassicType ------------------
# ----------------------------------------------


class Variable(BaseModel):
    """A type variable identified by a de Bruijn index."""

    t: Literal["V"] = "V"
    i: int
    b: "TypeBound"


class USize(BaseModel):
    """Unsigned integer size type."""

    t: Literal["I"] = "I"


class FunctionType(BaseModel):
    """A graph encoded as a value. It contains a concrete signature and a set of
    required resources."""

    input: "TypeRow"  # Value inputs of the function.
    output: "TypeRow"  # Value outputs of the function.
    # The extension requirements which are added by the operation
    extension_reqs: "ExtensionSet" = Field(default_factory=list)

    @classmethod
    def empty(cls) -> "FunctionType":
        return FunctionType(input=[], output=[], extension_reqs=[])


class PolyFuncType(BaseModel):
    """A graph encoded as a value. It contains a concrete signature and a set of
    required resources."""

    t: Literal["G"] = "G"

    # The declared type parameters, i.e., these must be instantiated with the same
    # number of TypeArgs before the function can be called. Note that within the body,
    # variable (DeBruijn) index 0 is element 0 of this array, i.e. the variables are
    # bound from right to left.
    params: list[TypeParam]

    # Template for the function. May contain variables up to length of `params`
    body: FunctionType

    @classmethod
    def empty(cls) -> "PolyFuncType":
        return PolyFuncType(params=[], body=FunctionType.empty())


class TypeBound(Enum):
    Eq = "E"
    Copyable = "C"
    Any = "A"

    @staticmethod
    def join(*bs: "TypeBound") -> "TypeBound":
        """Computes the least upper bound for a sequence of bounds."""
        res = TypeBound.Eq
        for b in bs:
            if b == TypeBound.Any:
                return TypeBound.Any
            if res == TypeBound.Eq:
                res = b
        return res


class Opaque(BaseModel):
    """An opaque operation that can be downcasted by the extensions that define it."""

    t: Literal["Opaque"] = "Opaque"
    extension: ExtensionId
    id: str  # Unique identifier of the opaque type.
    args: list[TypeArg]
    bound: TypeBound


# ----------------------------------------------
# --------------- LinearType -------------------
# ----------------------------------------------


class Qubit(BaseModel):
    """A qubit."""

    t: Literal["Q"] = "Q"


Type = TypeAliasType(
    "Type",
    Annotated[
        Qubit | Variable | USize | PolyFuncType | Array | TupleType | Sum | Opaque,
        Field(discriminator="t"),
        WrapValidator(_json_custom_error_validator),
    ],
)


# -------------------------------------------
# --------------- TypeRow -------------------
# -------------------------------------------

TypeRow = list[Type]


# -------------------------------------------
# --------------- Signature -----------------
# -------------------------------------------


class Signature(BaseModel):
    """Describes the edges required to/from a node.

    This includes both the concept of "signature" in the spec, and also the target
    (value) of a call (constant).
    """

    signature: "PolyFuncType"  # The underlying signature

    # The extensions which are associated with all the inputs and carried through
    input_extensions: ExtensionSet


# Now that all classes are defined, we need to update the ForwardRefs in all type
# annotations. We use some inspect magic to find all classes defined in this file.
classes = inspect.getmembers(
    sys.modules[__name__],
    lambda member: inspect.isclass(member) and member.__module__ == __name__,
)
for _, c in classes:
    if issubclass(c, BaseModel):
        c.model_rebuild()



import ast
import itertools
from abc import ABC, abstractmethod
from collections.abc import Iterator, Sequence
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Literal,
)

import guppylang.hugr.tys as tys
from guppylang.ast_util import AstNode

if TYPE_CHECKING:
    from guppylang.checker.core import Globals


Subst = dict["ExistentialTypeVar", "GuppyType"]
Inst = Sequence["GuppyType"]


@dataclass(frozen=True)
class GuppyType(ABC):
    """Base class for all Guppy types.

    Note that all instances of `GuppyType` subclasses are expected to be immutable.
    """

    name: ClassVar[str]

    # Cache for free variables
    _unsolved_vars: set["ExistentialTypeVar"] = field(init=False, repr=False)

    def __post_init__(self) -> None:
        # Make sure that we don't have higher-rank polymorphic types
        for arg in self.type_args:
            if isinstance(arg, FunctionType) and arg.quantified:
                from guppylang.error import InternalGuppyError

                raise InternalGuppyError(
                    "Tried to construct a higher-rank polymorphic type!"
                )

        # Compute free variables
        if isinstance(self, ExistentialTypeVar):
            vs = {self}
        else:
            vs = set()
            for arg in self.type_args:
                vs |= arg.unsolved_vars
        object.__setattr__(self, "_unsolved_vars", vs)

    @staticmethod
    @abstractmethod
    def build(*args: "GuppyType", node: AstNode | None = None) -> "GuppyType":
        pass

    @property
    @abstractmethod
    def type_args(self) -> Iterator["GuppyType"]:
        pass

    @property
    @abstractmethod
    def linear(self) -> bool:
        pass

    @abstractmethod
    def to_hugr(self) -> tys.Type:
        pass

    @abstractmethod
    def transform(self, transformer: "TypeTransformer") -> "GuppyType":
        pass

    def hugr_bound(self) -> tys.TypeBound:
        if self.linear:
            return tys.TypeBound.Any
        return tys.TypeBound.join(*(ty.hugr_bound() for ty in self.type_args))

    @property
    def unsolved_vars(self) -> set["ExistentialTypeVar"]:
        return self._unsolved_vars

    def substitute(self, s: Subst) -> "GuppyType":
        return self.transform(Substituter(s))


@dataclass(frozen=True)
class BoundTypeVar(GuppyType):
    """Bound type variable, identified with a de Bruijn index."""

    idx: int
    display_name: str
    linear: bool = False

    name: ClassVar[Literal["BoundTypeVar"]] = "BoundTypeVar"

    @staticmethod
    def build(*rgs: GuppyType, node: AstNode | None = None) -> GuppyType:
        raise NotImplementedError

    @property
    def type_args(self) -> Iterator["GuppyType"]:
        return iter(())

    def hugr_bound(self) -> tys.TypeBound:
        # We shouldn't make variables equatable, since we also want to substitute types
        # like `float`
        return tys.TypeBound.Any if self.linear else tys.TypeBound.Copyable

    def transform(self, transformer: "TypeTransformer") -> GuppyType:
        return transformer.transform(self) or self

    def __str__(self) -> str:
        return self.display_name

    def to_hugr(self) -> tys.Type:
        return tys.Variable(i=self.idx, b=self.hugr_bound())


@dataclass(frozen=True)
class ExistentialTypeVar(GuppyType):
    """Existential type variable, identified with a globally unique id.

    Is solved during type checking.
    """

    id: int
    display_name: str
    linear: bool = False

    name: ClassVar[Literal["ExistentialTypeVar"]] = "ExistentialTypeVar"

    _id_generator: ClassVar[Iterator[int]] = itertools.count()

    @classmethod
    def new(cls, display_name: str, linear: bool) -> "ExistentialTypeVar":
        return ExistentialTypeVar(next(cls._id_generator), display_name, linear)

    @staticmethod
    def build(*rgs: GuppyType, node: AstNode | None = None) -> GuppyType:
        raise NotImplementedError

    @property
    def type_args(self) -> Iterator["GuppyType"]:
        return iter(())

    def transform(self, transformer: "TypeTransformer") -> GuppyType:
        return transformer.transform(self) or self

    def __str__(self) -> str:
        return "?" + self.display_name

    def __hash__(self) -> int:
        return self.id

    def to_hugr(self) -> tys.Type:
        from guppylang.error import InternalGuppyError

        raise InternalGuppyError("Tried to convert free type variable to Hugr")


@dataclass(frozen=True)
class FunctionType(GuppyType):
    args: Sequence[GuppyType]
    returns: GuppyType
    arg_names: Sequence[str] | None = field(
        default=None,
        compare=False,  # Argument names are not taken into account for type equality
    )
    quantified: Sequence[BoundTypeVar] = field(default_factory=list)

    name: ClassVar[Literal["%function"]] = "%function"
    linear = False

    def __str__(self) -> str:
        prefix = (
            "forall " + ", ".join(str(v) for v in self.quantified) + ". "
            if self.quantified
            else ""
        )
        if len(self.args) == 1:
            [arg] = self.args
            return prefix + f"{arg} -> {self.returns}"
        else:
            return (
                prefix + f"({', '.join(str(a) for a in self.args)}) -> {self.returns}"
            )

    @staticmethod
    def build(*args: GuppyType, node: AstNode | None = None) -> GuppyType:
        # Function types cannot be constructed using `build`. The type parsing code
        # has a special case for function types.
        raise NotImplementedError

    @property
    def type_args(self) -> Iterator[GuppyType]:
        return itertools.chain(iter(self.args), iter((self.returns,)))

    def to_hugr(self) -> tys.PolyFuncType:
        ins = [t.to_hugr() for t in self.args]
        outs = [t.to_hugr() for t in type_to_row(self.returns)]
        func_ty = tys.FunctionType(input=ins, output=outs, extension_reqs=[])
        return tys.PolyFuncType(
            params=[tys.TypeTypeParam(b=v.hugr_bound()) for v in self.quantified],
            body=func_ty,
        )

    def hugr_bound(self) -> tys.TypeBound:
        # Functions are not equatable, only copyable
        return tys.TypeBound.Copyable

    def transform(self, transformer: "TypeTransformer") -> GuppyType:
        return transformer.transform(self) or FunctionType(
            [ty.transform(transformer) for ty in self.args],
            self.returns.transform(transformer),
            self.arg_names,
        )

    def instantiate(self, tys: Sequence[GuppyType]) -> "FunctionType":
        """Instantiates quantified type variables."""
        assert len(tys) == len(self.quantified)

        # Set the `preserve` flag for instantiated tuples and None
        preserved_tys: list[GuppyType] = []
        for ty in tys:
            if isinstance(ty, TupleType):
                ty = TupleType(ty.element_types, preserve=True)
            elif isinstance(ty, NoneType):
                ty = NoneType(preserve=True)
            preserved_tys.append(ty)

        inst = Instantiator(preserved_tys)
        return FunctionType(
            [ty.transform(inst) for ty in self.args],
            self.returns.transform(inst),
            self.arg_names,
        )

    def unquantified(self) -> tuple["FunctionType", Sequence[ExistentialTypeVar]]:
        """Replaces all quantified variables with free type variables."""
        inst = [
            ExistentialTypeVar.new(v.display_name, v.linear) for v in self.quantified
        ]
        return self.instantiate(inst), inst


@dataclass(frozen=True)
class TupleType(GuppyType):
    element_types: Sequence[GuppyType]

    # Flag to avoid turning the tuple into row when calling `type_to_row()`. This is
    # used to make sure that type vars instantiated to tuples are not broken up into
    # rows when generating a Hugr
    preserve: bool = field(default=False, compare=False)

    name: ClassVar[Literal["tuple"]] = "tuple"

    @staticmethod
    def build(*args: GuppyType, node: AstNode | None = None) -> GuppyType:
        from guppylang.error import GuppyError

        # TODO: Parse empty tuples via `tuple[()]`
        if len(args) == 0:
            raise GuppyError("Tuple type requires generic type arguments", node)
        return TupleType(list(args))

    def __str__(self) -> str:
        return f"({', '.join(str(e) for e in self.element_types)})"

    @property
    def linear(self) -> bool:
        return any(t.linear for t in self.element_types)

    @property
    def type_args(self) -> Iterator[GuppyType]:
        return iter(self.element_types)

    def to_hugr(self) -> tys.Type:
        ts = [t.to_hugr() for t in self.element_types]
        return tys.TupleType(inner=ts)

    def transform(self, transformer: "TypeTransformer") -> GuppyType:
        return transformer.transform(self) or TupleType(
            [ty.transform(transformer) for ty in self.element_types]
        )


@dataclass(frozen=True)
class SumType(GuppyType):
    element_types: Sequence[GuppyType]

    name: ClassVar[str] = "%tuple"

    @staticmethod
    def build(*args: GuppyType, node: AstNode | None = None) -> GuppyType:
        # Sum types cannot be parsed and constructed using `build` since they cannot be
        # written by the user
        raise NotImplementedError

    def __str__(self) -> str:
        return f"Sum({', '.join(str(e) for e in self.element_types)})"

    @property
    def linear(self) -> bool:
        return any(t.linear for t in self.element_types)

    @property
    def type_args(self) -> Iterator[GuppyType]:
        return iter(self.element_types)

    def to_hugr(self) -> tys.Type:
        if all(
            isinstance(e, TupleType) and len(e.element_types) == 0
            for e in self.element_types
        ):
            return tys.UnitSum(size=len(self.element_types))
        return tys.GeneralSum(row=[t.to_hugr() for t in self.element_types])

    def transform(self, transformer: "TypeTransformer") -> GuppyType:
        return transformer.transform(self) or SumType(
            [ty.transform(transformer) for ty in self.element_types]
        )


@dataclass(frozen=True)
class ListType(GuppyType):
    element_type: GuppyType

    name: ClassVar[Literal["list"]] = "list"
    linear: bool = field(default=False, init=False)

    @staticmethod
    def build(*args: GuppyType, node: AstNode | None = None) -> GuppyType:
        from guppylang.error import GuppyError

        if len(args) == 0:
            raise GuppyError("Missing type parameter for generic type `list`", node)
        if len(args) > 1:
            raise GuppyError("Too many type arguments for generic type `list`", node)
        (arg,) = args
        if arg.linear:
            raise GuppyError(
                "Type `list` cannot store linear data, use `linst` instead", node
            )
        return ListType(arg)

    def __str__(self) -> str:
        return f"list[{self.element_type}]"

    @property
    def type_args(self) -> Iterator[GuppyType]:
        return iter((self.element_type,))

    def to_hugr(self) -> tys.Type:
        return tys.Opaque(
            extension="Collections",
            id="List",
            args=[tys.TypeTypeArg(ty=self.element_type.to_hugr())],
            bound=self.hugr_bound(),
        )

    def transform(self, transformer: "TypeTransformer") -> GuppyType:
        return transformer.transform(self) or ListType(
            self.element_type.transform(transformer)
        )


@dataclass(frozen=True)
class LinstType(GuppyType):
    element_type: GuppyType

    name: ClassVar[Literal["linst"]] = "linst"

    @staticmethod
    def build(*args: GuppyType, node: AstNode | None = None) -> GuppyType:
        from guppylang.error import GuppyError

        if len(args) == 0:
            raise GuppyError("Missing type parameter for generic type `linst`", node)
        if len(args) > 1:
            raise GuppyError("Too many type arguments for generic type `linst`", node)
        return LinstType(args[0])

    def __str__(self) -> str:
        return f"linst[{self.element_type}]"

    @property
    def linear(self) -> bool:
        return self.element_type.linear

    @property
    def type_args(self) -> Iterator[GuppyType]:
        return iter((self.element_type,))

    def to_hugr(self) -> tys.Type:
        return tys.Opaque(
            extension="Collections",
            id="List",
            args=[tys.TypeTypeArg(ty=self.element_type.to_hugr())],
            bound=self.hugr_bound(),
        )

    def transform(self, transformer: "TypeTransformer") -> GuppyType:
        return transformer.transform(self) or LinstType(
            self.element_type.transform(transformer)
        )


@dataclass(frozen=True)
class NoneType(GuppyType):
    name: ClassVar[Literal["None"]] = "None"
    linear: bool = False

    # Flag to avoid turning the type into a row when calling `type_to_row()`. This is
    # used to make sure that type vars instantiated to Nones are not broken up into
    # empty rows when generating a Hugr
    preserve: bool = field(default=False, compare=False)

    @staticmethod
    def build(*args: GuppyType, node: AstNode | None = None) -> GuppyType:
        if len(args) > 0:
            from guppylang.error import GuppyError

            raise GuppyError("Type `None` is not generic", node)
        return NoneType()

    @property
    def type_args(self) -> Iterator[GuppyType]:
        return iter(())

    def substitute(self, s: Subst) -> GuppyType:
        return self

    def __str__(self) -> str:
        return "None"

    def to_hugr(self) -> tys.Type:
        return tys.TupleType(inner=[])

    def transform(self, transformer: "TypeTransformer") -> GuppyType:
        return transformer.transform(self) or self


@dataclass(frozen=True)
class BoolType(SumType):
    """The type of booleans."""

    linear: bool = False
    name: ClassVar[Literal["bool"]] = "bool"

    def __init__(self) -> None:
        # Hugr bools are encoded as Sum((), ())
        super().__init__([TupleType([]), TupleType([])])

    @staticmethod
    def build(*args: GuppyType, node: AstNode | None = None) -> GuppyType:
        if len(args) > 0:
            from guppylang.error import GuppyError

            raise GuppyError("Type `bool` is not generic", node)
        return BoolType()

    def __str__(self) -> str:
        return "bool"

    def transform(self, transformer: "TypeTransformer") -> GuppyType:
        return transformer.transform(self) or self


class TypeTransformer(ABC):
    """Abstract base class for a type visitor that transforms types."""

    @abstractmethod
    def transform(self, ty: GuppyType) -> GuppyType | None:
        """This method is called for each visited type.

        Return a transformed type or `None` to continue the recursive visit.
        """


class Substituter(TypeTransformer):
    """Type transformer that substitutes free type variables."""

    subst: Subst

    def __init__(self, subst: Subst) -> None:
        self.subst = subst

    def transform(self, ty: GuppyType) -> GuppyType | None:
        if isinstance(ty, ExistentialTypeVar):
            return self.subst.get(ty, None)
        return None


class Instantiator(TypeTransformer):
    """Type transformer that instantiates bound type variables."""

    tys: Sequence[GuppyType]

    def __init__(self, tys: Sequence[GuppyType]) -> None:
        self.tys = tys

    def transform(self, ty: GuppyType) -> GuppyType | None:
        if isinstance(ty, BoundTypeVar):
            # Instantiate if type for the index is available
            if ty.idx < len(self.tys):
                return self.tys[ty.idx]

            # Otherwise, lower the de Bruijn index
            return BoundTypeVar(ty.idx - len(self.tys), ty.display_name, ty.linear)
        return None


def unify(s: GuppyType, t: GuppyType, subst: Subst | None) -> Subst | None:
    """Computes a most general unifier for two types.

    Return a substitutions `subst` such that `s[subst] == t[subst]` or `None` if this
    not possible.
    """
    if subst is None:
        return None
    if s == t:
        return subst
    if isinstance(s, ExistentialTypeVar):
        return _unify_var(s, t, subst)
    if isinstance(t, ExistentialTypeVar):
        return _unify_var(t, s, subst)
    if type(s) == type(t):
        sargs, targs = list(s.type_args), list(t.type_args)
        if len(sargs) == len(targs):
            for sa, ta in zip(sargs, targs):
                subst = unify(sa, ta, subst)
            return subst
    return None


def _unify_var(var: ExistentialTypeVar, t: GuppyType, subst: Subst) -> Subst | None:
    """Helper function for unification of type variables."""
    if var in subst:
        return unify(subst[var], t, subst)
    if isinstance(t, ExistentialTypeVar) and t in subst:
        return unify(var, subst[t], subst)
    if var in t.unsolved_vars:
        return None
    return {var: t, **subst}


def type_from_ast(
    node: AstNode,
    globals: "Globals",
    type_var_mapping: dict[str, BoundTypeVar] | None = None,
) -> GuppyType:
    """Turns an AST expression into a Guppy type."""
    from guppylang.error import GuppyError

    if isinstance(node, ast.Name):
        x = node.id
        if x in globals.types:
            return globals.types[x].build(node=node)
        if x in globals.type_vars:
            if type_var_mapping is None:
                raise GuppyError(
                    "Free type variable. Only function types can be generic", node
                )
            var_decl = globals.type_vars[x]
            if var_decl.name not in type_var_mapping:
                type_var_mapping[var_decl.name] = BoundTypeVar(
                    len(type_var_mapping), var_decl.name, var_decl.linear
                )
            return type_var_mapping[var_decl.name]
        raise GuppyError("Unknown type", node)

    if isinstance(node, ast.Constant):
        v = node.value
        if v is None:
            return NoneType()
        if isinstance(v, str):
            try:
                [stmt] = ast.parse(v).body
                if not isinstance(stmt, ast.Expr):
                    raise GuppyError("Invalid Guppy type", node)
                return type_from_ast(stmt.value, globals, type_var_mapping)
            except (SyntaxError, ValueError):
                raise GuppyError("Invalid Guppy type", node) from None
        raise GuppyError(f"Constant `{v}` is not a valid type", node)

    if isinstance(node, ast.Tuple):
        return TupleType(
            [type_from_ast(el, globals, type_var_mapping) for el in node.elts]
        )

    if (
        isinstance(node, ast.Subscript)
        and isinstance(node.value, ast.Name)
        and node.value.id == "Callable"
        and isinstance(node.slice, ast.Tuple)
        and len(node.slice.elts) == 2
    ):
        # TODO: Do we want to allow polymorphic Callable types?
        [func_args, ret] = node.slice.elts
        if isinstance(func_args, ast.List):
            return FunctionType(
                [type_from_ast(a, globals, type_var_mapping) for a in func_args.elts],
                type_from_ast(ret, globals, type_var_mapping),
            )

    if isinstance(node, ast.Subscript) and isinstance(node.value, ast.Name):
        x = node.value.id
        if x in globals.types:
            args = (
                node.slice.elts if isinstance(node.slice, ast.Tuple) else [node.slice]
            )
            return globals.types[x].build(
                *(type_from_ast(a, globals, type_var_mapping) for a in args), node=node
            )

    raise GuppyError("Not a valid Guppy type", node)


def type_row_from_ast(node: ast.expr, globals: "Globals") -> Sequence[GuppyType]:
    """Turns an AST expression into a Guppy type row.

    This is needed to interpret the return type annotation of functions.
    """
    # The return type `-> None` is represented in the ast as `ast.Constant(value=None)`
    if isinstance(node, ast.Constant) and node.value is None:
        return []
    ty = type_from_ast(node, globals)
    if isinstance(ty, TupleType):
        return ty.element_types
    else:
        return [ty]


def row_to_type(row: Sequence[GuppyType]) -> GuppyType:
    """Turns a row of types into a single type by packing into a tuple."""
    if len(row) == 0:
        return NoneType()
    elif len(row) == 1:
        return row[0]
    else:
        return TupleType(row)


def type_to_row(ty: GuppyType) -> Sequence[GuppyType]:
    """Turns a type into a row of types by unpacking top-level tuples."""
    if isinstance(ty, NoneType) and not ty.preserve:
        return []
    if isinstance(ty, TupleType) and not ty.preserve:
        return ty.element_types
    return [ty]



import ast
from abc import ABC, abstractmethod

from guppylang.ast_util import AstNode, get_type, with_loc, with_type
from guppylang.checker.core import Context, Globals
from guppylang.checker.expr_checker import check_call, synthesize_call
from guppylang.checker.func_checker import check_signature
from guppylang.compiler.core import CompiledFunction, CompiledGlobals, DFContainer
from guppylang.error import (
    GuppyError,
    InternalGuppyError,
    UnknownFunctionType,
)
from guppylang.gtypes import FunctionType, GuppyType, Inst, Subst, type_to_row
from guppylang.hugr import ops
from guppylang.hugr.hugr import DFContainingVNode, Hugr, Node, OutPortV
from guppylang.nodes import GlobalCall


class CustomFunction(CompiledFunction):
    """A function whose type checking and compilation behaviour can be customised."""

    defined_at: ast.FunctionDef | None

    # Whether the function may be used as a higher-order value. This is only possible
    # if a static type for the function is provided.
    higher_order_value: bool

    call_checker: "CustomCallChecker"
    call_compiler: "CustomCallCompiler"

    _ty: FunctionType | None = None
    _defined: dict[Node, DFContainingVNode] = {}  # noqa: RUF012

    def __init__(
        self,
        name: str,
        defined_at: ast.FunctionDef | None,
        compiler: "CustomCallCompiler",
        checker: "CustomCallChecker",
        higher_order_value: bool = True,
        ty: FunctionType | None = None,
    ):
        self.name = name
        self.defined_at = defined_at
        self.higher_order_value = higher_order_value
        self.call_compiler = compiler
        self.call_checker = checker
        self.used = None
        self._ty = ty
        self._defined = {}

    @property  # type: ignore[override]
    def ty(self) -> FunctionType:
        if self._ty is None:
            return UnknownFunctionType()
        return self._ty

    @ty.setter
    def ty(self, ty: FunctionType) -> None:
        self._ty = ty

    def check_type(self, globals: Globals) -> None:
        """Checks the type annotation on the signature declaration if provided."""
        if self._ty is not None:
            return

        if self.defined_at is None:
            if self.higher_order_value:
                raise GuppyError(
                    f"Type signature for function `{self.name}` is required. "
                    "Alternatively, try passing `higher_order_value=False` on "
                    "definition."
                )
            return

        try:
            self._ty = check_signature(self.defined_at, globals)
        except GuppyError:
            # We can ignore the error if a custom call checker is provided and the
            # function may not be used as a higher-order value
            if self.call_checker is None or self.higher_order_value:
                raise

    def check_call(
        self, args: list[ast.expr], ty: GuppyType, node: AstNode, ctx: Context
    ) -> tuple[ast.expr, Subst]:
        self.call_checker._setup(ctx, node, self)
        new_node, subst = self.call_checker.check(args, ty)
        return with_type(ty, with_loc(node, new_node)), subst

    def synthesize_call(
        self, args: list[ast.expr], node: AstNode, ctx: "Context"
    ) -> tuple[ast.expr, GuppyType]:
        self.call_checker._setup(ctx, node, self)
        new_node, ty = self.call_checker.synthesize(args)
        return with_type(ty, with_loc(node, new_node)), ty

    def compile_call(
        self,
        args: list[OutPortV],
        type_args: Inst,
        dfg: DFContainer,
        graph: Hugr,
        globals: CompiledGlobals,
        node: AstNode,
    ) -> list[OutPortV]:
        self.call_compiler._setup(type_args, dfg, graph, globals, node)
        return self.call_compiler.compile(args)

    def load(
        self, dfg: "DFContainer", graph: Hugr, globals: CompiledGlobals, node: AstNode
    ) -> OutPortV:
        """Loads the custom function as a value into a local dataflow graph.

        This will place a `FunctionDef` node into the Hugr module if one for this
        function doesn't already exist and loads it into the DFG. This operation will
        fail if no function type has been specified.
        """
        if self._ty is None:
            raise GuppyError(
                "This function does not support usage in a higher-order context",
                node,
            )

        if self._ty.quantified:
            raise InternalGuppyError(
                "Can't yet generate higher-order versions of custom functions. This "
                "requires generic function *definitions*"
            )

        # Find the module node by walking up the hierarchy
        module: Node = dfg.node
        while not isinstance(module.op, ops.Module):
            if module.parent is None:
                raise InternalGuppyError(
                    "Encountered node that is not contained in a module."
                )
            module = module.parent

        # If the function has not yet been loaded in this module, we first have to
        # define it. We create a `FunctionDef` that takes some inputs, compiles a call
        # to the function, and returns the results.
        if module not in self._defined:
            def_node = graph.add_def(self.ty, module, self.name)
            _, inp_ports = graph.add_input_with_ports(list(self.ty.args), def_node)
            returns = self.compile_call(
                inp_ports, [], DFContainer(def_node, {}), graph, globals, node
            )
            graph.add_output(returns, parent=def_node)
            self._defined[module] = def_node

        # Finally, load the function into the local DFG
        return graph.add_load_constant(
            self._defined[module].out_port(0), dfg.node
        ).out_port(0)


class CustomCallChecker(ABC):
    """Protocol for custom function call type checkers."""

    ctx: Context
    node: AstNode
    func: CustomFunction

    def _setup(self, ctx: Context, node: AstNode, func: CustomFunction) -> None:
        self.ctx = ctx
        self.node = node
        self.func = func

    @abstractmethod
    def check(self, args: list[ast.expr], ty: GuppyType) -> tuple[ast.expr, Subst]:
        """Checks the return value against a given type.

        Returns a (possibly) transformed and annotated AST node for the call.
        """

    @abstractmethod
    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, GuppyType]:
        """Synthesizes a type for the return value of a call.

        Also returns a (possibly) transformed and annotated argument list.
        """


class CustomCallCompiler(ABC):
    """Protocol for custom function call compilers."""

    type_args: Inst
    dfg: DFContainer
    graph: Hugr
    globals: CompiledGlobals
    node: AstNode

    def _setup(
        self,
        type_args: Inst,
        dfg: DFContainer,
        graph: Hugr,
        globals: CompiledGlobals,
        node: AstNode,
    ) -> None:
        self.type_args = type_args
        self.dfg = dfg
        self.graph = graph
        self.globals = globals
        self.node = node

    @abstractmethod
    def compile(self, args: list[OutPortV]) -> list[OutPortV]:
        """Compiles a custom function call and returns the resulting ports."""


class DefaultCallChecker(CustomCallChecker):
    """Checks function calls by comparing to a type signature."""

    def check(self, args: list[ast.expr], ty: GuppyType) -> tuple[ast.expr, Subst]:
        # Use default implementation from the expression checker
        args, subst, inst = check_call(self.func.ty, args, ty, self.node, self.ctx)
        return GlobalCall(func=self.func, args=args, type_args=inst), subst

    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, GuppyType]:
        # Use default implementation from the expression checker
        args, ty, inst = synthesize_call(self.func.ty, args, self.node, self.ctx)
        return GlobalCall(func=self.func, args=args, type_args=inst), ty


class DefaultCallCompiler(CustomCallCompiler):
    """Call compiler that invokes the regular expression compiler."""

    def compile(self, args: list[OutPortV]) -> list[OutPortV]:
        raise NotImplementedError


class OpCompiler(CustomCallCompiler):
    op: ops.BaseOp

    def __init__(self, op: ops.BaseOp) -> None:
        self.op = op

    def compile(self, args: list[OutPortV]) -> list[OutPortV]:
        node = self.graph.add_node(
            self.op.model_copy(), inputs=args, parent=self.dfg.node
        )
        return_ty = get_type(self.node)
        return [node.add_out_port(ty) for ty in type_to_row(return_ty)]


class NoopCompiler(CustomCallCompiler):
    def compile(self, args: list[OutPortV]) -> list[OutPortV]:
        return args



import inspect
import sys
from abc import ABC
from typing import Annotated, Any, Literal, cast

from pydantic import BaseModel, Field
from typing_extensions import TypeAliasType

import guppylang.hugr.tys as tys

from .tys import (
    ExtensionId,
    ExtensionSet,
    FunctionType,
    PolyFuncType,
    Type,
    TypeRow,
)
from .val import Value

NodeID = int


class BaseOp(ABC, BaseModel):
    """Base class for ops that store their node's input/output types"""

    # Parent node index of node the op belongs to, used only at serialization time
    parent: NodeID = 0
    input_extensions: ExtensionSet | None = None

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        """Hook to insert type information from the input and output ports into the
        op"""

    def insert_child_dfg_signature(self, inputs: TypeRow, outputs: TypeRow) -> None:
        """Hook to insert type information from a child dataflow graph"""

    def display_name(self) -> str:
        """Name of the op for visualisation"""
        return self.__class__.__name__


class DummyOp(BaseOp):
    """Nodes used inside dataflow containers (DFG, Conditional, TailLoop, def,
    BasicBlock)."""

    op: Literal["DummyOp"] = "DummyOp"
    name: str

    def display_name(self) -> str:
        return f'"{self.name}"'


# ----------------------------------------------------------
# --------------- Module level operations ------------------
# ----------------------------------------------------------


class Module(BaseOp):
    """The root of a module, parent of all other `ModuleOp`s."""

    op: Literal["Module"] = "Module"


class FuncDefn(BaseOp):
    """A function definition. Children nodes are the body of the definition."""

    op: Literal["FuncDefn"] = "FuncDefn"

    name: str
    signature: PolyFuncType = Field(default_factory=PolyFuncType.empty)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        assert len(in_types) == 0
        assert len(out_types) == 1
        out = out_types[0]
        assert isinstance(out, PolyFuncType)
        self.signature = out  # TODO: Extensions


class FuncDecl(BaseOp):
    """External function declaration, linked at runtime."""

    op: Literal["FuncDecl"] = "FuncDecl"
    name: str
    signature: PolyFuncType = Field(default_factory=PolyFuncType.empty)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        assert len(in_types) == 0
        assert len(out_types) == 1
        out = out_types[0]
        assert isinstance(out, PolyFuncType)
        self.signature = out


class Const(BaseOp):
    """A constant value definition."""

    op: Literal["Const"] = "Const"
    value: Value
    typ: Type


# -----------------------------------------------
# --------------- BasicBlock types ------------------
# -----------------------------------------------


class DataflowBlock(BaseOp):
    """A CFG basic block node. The signature is that of the internal Dataflow
    graph."""

    op: Literal["DataflowBlock"] = "DataflowBlock"
    inputs: TypeRow = Field(default_factory=list)
    other_outputs: TypeRow = Field(default_factory=list)
    tuple_sum_rows: list[TypeRow] = Field(default_factory=list)
    extension_delta: ExtensionSet = Field(default_factory=list)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        # The types will be all None because it's not dataflow, but we only care about
        # the number of outputs. Note that we don't make use of the HUGR feature where
        # the variant data is appended to successor input. Thus, `predicate_variants`
        # will only contain empty rows.
        num_cases = len(out_types)
        self.tuple_sum_rows = [[] for _ in range(num_cases)]

    def insert_child_dfg_signature(self, inputs: TypeRow, outputs: TypeRow) -> None:
        self.inputs = inputs
        pred = outputs[0]
        assert isinstance(pred, tys.UnitSum | tys.GeneralSum)
        if isinstance(pred, tys.UnitSum):
            self.tuple_sum_rows = [[] for _ in range(cast(tys.UnitSum, pred).size)]
        else:
            assert isinstance(pred, tys.GeneralSum)
            self.tuple_sum_rows = []
            for variant in pred.row:
                assert isinstance(variant, tys.TupleType)
                self.tuple_sum_rows.append(variant.inner)
        self.other_outputs = outputs[1:]


class ExitBlock(BaseOp):
    """The single exit node of the CFG, has no children, stores the types of
    the CFG node output."""

    op: Literal["ExitBlock"] = "ExitBlock"
    cfg_outputs: TypeRow


# ---------------------------------------------
# --------------- DataflowOp ------------------
# ---------------------------------------------


class DataflowOp(BaseOp):
    pass


class Input(DataflowOp):
    """An input node. The outputs of this node are the inputs to the function."""

    op: Literal["Input"] = "Input"
    types: TypeRow = Field(default_factory=list)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        assert len(in_types) == 0
        self.types = list(out_types)


class Output(DataflowOp):
    """An output node. The inputs are the outputs of the function."""

    op: Literal["Output"] = "Output"
    types: TypeRow = Field(default_factory=list)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        assert len(out_types) == 0
        self.types = list(in_types)


class Call(DataflowOp):
    """
    Call a function directly.

    The first port is connected to the def/declare of the function being called
    directly, with a `ConstE<Graph>` edge. The signature of the remaining ports matches
    the function being called.
    """

    op: Literal["Call"] = "Call"
    signature: FunctionType = Field(default_factory=FunctionType.empty)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        # The constE edge comes after the value inputs
        fun_ty = in_types[-1]
        assert isinstance(fun_ty, PolyFuncType)
        fun_ty = cast(PolyFuncType, fun_ty)
        assert len(fun_ty.params) == 0
        self.signature = fun_ty.body


class CallIndirect(DataflowOp):
    """Call a function indirectly.

    Like call, but the first input is a standard dataflow graph type."""

    op: Literal["CallIndirect"] = "CallIndirect"
    signature: FunctionType = Field(default_factory=FunctionType.empty)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        fun_ty = in_types[0]
        assert isinstance(fun_ty, PolyFuncType)
        fun_ty = cast(PolyFuncType, fun_ty)
        assert len(fun_ty.params) == 0
        assert len(fun_ty.body.input) == len(in_types) - 1
        assert len(fun_ty.body.output) == len(out_types)
        self.signature = fun_ty.body


class LoadConstant(DataflowOp):
    """Load a static constant in to the local dataflow graph."""

    op: Literal["LoadConstant"] = "LoadConstant"
    datatype: Type


class LeafOp(DataflowOp):
    """Simple operation that has only value inputs+outputs and (potentially) StateOrder
    edges."""

    op: Literal["LeafOp"] = "LeafOp"


class DFG(DataflowOp):
    """A simply nested dataflow graph."""

    op: Literal["DFG"] = "DFG"
    signature: FunctionType = Field(default_factory=FunctionType.empty)

    def insert_child_dfg_signature(self, inputs: TypeRow, outputs: TypeRow) -> None:
        self.signature = FunctionType(
            input=list(inputs), output=list(outputs), extension_reqs=[]
        )


# ------------------------------------------------
# --------------- ControlFlowOp ------------------
# ------------------------------------------------


class Conditional(DataflowOp):
    """Conditional operation, defined by child `Case` nodes for each branch."""

    op: Literal["Conditional"] = "Conditional"
    tuple_sum_rows: list[TypeRow] = Field(
        default_factory=list
    )  # The possible rows of the predicate input
    other_inputs: TypeRow = Field(default_factory=list)  # Remaining input types
    outputs: TypeRow = Field(default_factory=list)  # Output types
    # Extensions used to produce the outputs
    extension_delta: ExtensionSet = Field(default_factory=list)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        # First port is a predicate, i.e. a sum of tuple types. We need to unpack
        # those into a list of type rows
        pred = in_types[0]
        if isinstance(pred, tys.UnitSum):
            self.tuple_sum_rows = [[] for _ in range(cast(tys.UnitSum, pred).size)]
        else:
            assert isinstance(pred, tys.GeneralSum)
            self.tuple_sum_rows = []
            for ty in pred.row:
                assert isinstance(ty, tys.TupleType)
                self.tuple_sum_rows.append(ty.inner)
        self.other_inputs = list(in_types[1:])
        self.outputs = list(out_types)


class Case(BaseOp):
    """Case ops - nodes valid inside Conditional nodes."""

    op: Literal["Case"] = "Case"
    # The signature of the contained dataflow graph.
    signature: FunctionType = Field(default_factory=FunctionType.empty)

    def insert_child_dfg_signature(self, inputs: TypeRow, outputs: TypeRow) -> None:
        self.signature = tys.FunctionType(
            input=list(inputs), output=list(outputs), extension_reqs=[]
        )


class TailLoop(DataflowOp):
    """Tail-controlled loop."""

    op: Literal["TailLoop"] = "TailLoop"
    just_inputs: TypeRow = Field(default_factory=list)  # Types that are only input
    just_outputs: TypeRow = Field(default_factory=list)  # Types that are only output
    # Types that are appended to both input and output:
    rest: TypeRow = Field(default_factory=list)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        assert in_types == out_types
        # self.just_inputs = list(in_types)
        # self.just_outputs = list(out_types)
        self.rest = list(in_types)


class CFG(DataflowOp):
    """A dataflow node which is defined by a child CFG."""

    op: Literal["CFG"] = "CFG"
    signature: FunctionType = Field(default_factory=FunctionType.empty)

    def insert_port_types(self, inputs: TypeRow, outputs: TypeRow) -> None:
        self.signature = FunctionType(
            input=list(inputs), output=list(outputs), extension_reqs=[]
        )


ControlFlowOp = Conditional | TailLoop | CFG


# -----------------------------------------
# --------------- LeafOp ------------------
# -----------------------------------------


class CustomOp(LeafOp):
    """A user-defined operation that can be downcasted by the extensions that define
    it."""

    lop: Literal["CustomOp"] = "CustomOp"
    extension: ExtensionId
    op_name: str
    signature: tys.FunctionType = Field(default_factory=tys.FunctionType.empty)
    description: str = ""
    args: list[tys.TypeArg] = Field(default_factory=list)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        self.signature = tys.FunctionType(input=list(in_types), output=list(out_types))

    def display_name(self) -> str:
        return self.op_name


class Noop(LeafOp):
    """A no-op operation."""

    lop: Literal["Noop"] = "Noop"
    ty: Type

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        assert len(in_types) == 1
        assert len(out_types) == 1
        assert in_types[0] == out_types[0]  # type: ignore[operator]  # mypy doesn't understand the type union
        self.ty = in_types[0]


class MakeTuple(LeafOp):
    """An operation that packs all its inputs into a tuple."""

    lop: Literal["MakeTuple"] = "MakeTuple"
    tys: TypeRow = Field(default_factory=list)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        # If we have a single order edge as input, this is a unit
        if in_types == [None]:
            in_types = []
        self.tys = list(in_types)


class UnpackTuple(LeafOp):
    """An operation that packs all its inputs into a tuple."""

    lop: Literal["UnpackTuple"] = "UnpackTuple"
    tys: TypeRow = Field(default_factory=list)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        self.tys = list(out_types)


class Tag(LeafOp):
    """An operation that creates a tagged sum value from one of its variants."""

    lop: Literal["Tag"] = "Tag"
    tag: int  # The variant to create.
    variants: TypeRow  # The variants of the sum type.


class TypeApply(LeafOp):
    """Fixes some TypeParams of a polymorphic type by providing TypeArgs"""

    lop: Literal["TypeApply"] = "TypeApply"
    ta: "TypeApplication"


class TypeApplication(BaseModel):
    """Records details of an application of a PolyFuncType to some TypeArgs and the
    result (a less-, but still potentially-, polymorphic type).

    Note that Guppy only generates full type applications, where the result is a
    monomorphic type. Partial type applications are not used by Guppy.
    """

    input: PolyFuncType
    args: list[tys.TypeTypeArg]
    output: PolyFuncType


LeafOpUnion = TypeAliasType(
    "LeafOpUnion",
    Annotated[
        (CustomOp | Noop | MakeTuple | UnpackTuple | Tag | TypeApply),
        Field(discriminator="lop"),
    ],
)


OpType = TypeAliasType(
    "OpType",
    Annotated[
        (
            Module
            | Case
            | Module
            | FuncDefn
            | FuncDecl
            | Const
            | DummyOp
            | DataflowBlock
            | ExitBlock
            | Conditional
            | TailLoop
            | CFG
            | Input
            | Output
            | Call
            | CallIndirect
            | LoadConstant
            | LeafOpUnion
            | DFG
        ),
        Field(discriminator="op"),
    ],
)


# --------------------------------------
# --------------- OpDef ----------------
# --------------------------------------


class OpDef(BaseOp, populate_by_name=True):
    """Serializable definition for dynamically loaded operations."""

    name: str  # Unique identifier of the operation.
    description: str  # Human readable description of the operation.
    inputs: list[tuple[str | None, Type]]
    outputs: list[tuple[str | None, Type]]
    misc: dict[str, Any]  # Miscellaneous data associated with the operation.
    def_: str | None = Field(
        ..., alias="def"
    )  # (YAML?)-encoded definition of the operation.
    extension_reqs: ExtensionSet  # Resources required to execute this operation.


# Now that all classes are defined, we need to update the ForwardRefs in all type
# annotations. We use some inspect magic to find all classes defined in this file.
classes = inspect.getmembers(
    sys.modules[__name__],
    lambda member: inspect.isclass(member) and member.__module__ == __name__,
)
for _, c in classes:
    if issubclass(c, BaseModel):
        c.model_rebuild()



"""Guppy module for builtin types and operations."""

# mypy: disable-error-code="empty-body, misc, override, valid-type, no-untyped-def"

from guppylang.custom import DefaultCallChecker, NoopCompiler
from guppylang.decorator import guppy
from guppylang.gtypes import BoolType, LinstType, ListType
from guppylang.hugr import ops, tys
from guppylang.module import GuppyModule
from guppylang.prelude._internal import (
    CallableChecker,
    CoercingChecker,
    DunderChecker,
    FailingChecker,
    FloatBoolCompiler,
    FloatDivmodCompiler,
    FloatFloordivCompiler,
    FloatModCompiler,
    IntTruedivCompiler,
    ReversingChecker,
    UnsupportedChecker,
    float_op,
    hugr_float_type,
    hugr_int_type,
    int_op,
    logic_op,
)

builtins = GuppyModule("builtins", import_builtins=False)

T = guppy.type_var(builtins, "T")
L = guppy.type_var(builtins, "L", linear=True)


@guppy.extend_type(builtins, BoolType)
class Bool:
    @guppy.hugr_op(builtins, logic_op("And", [tys.BoundedNatArg(n=2)]))
    def __and__(self: bool, other: bool) -> bool:
        ...

    @guppy.custom(builtins, NoopCompiler())
    def __bool__(self: bool) -> bool:
        ...

    @guppy.hugr_op(builtins, int_op("ifrombool"))
    def __int__(self: bool) -> int:
        ...

    @guppy.hugr_op(builtins, logic_op("Or", [tys.BoundedNatArg(n=2)]))
    def __or__(self: bool, other: bool) -> bool:
        ...


@guppy.type(builtins, hugr_int_type, name="int")
class Int:
    @guppy.hugr_op(builtins, int_op("iabs"))  # TODO: Maybe wrong? (signed vs unsigned!)
    def __abs__(self: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("iadd"))
    def __add__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("iand"))
    def __and__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("itobool"))
    def __bool__(self: int) -> bool:
        ...

    @guppy.custom(builtins, NoopCompiler())
    def __ceil__(self: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("idivmod_s", num_params=2))
    def __divmod__(self: int, other: int) -> tuple[int, int]:
        ...

    @guppy.hugr_op(builtins, int_op("ieq"))
    def __eq__(self: int, other: int) -> bool:
        ...

    @guppy.hugr_op(builtins, int_op("convert_s", "arithmetic.conversions"))
    def __float__(self: int) -> float:
        ...

    @guppy.custom(builtins, NoopCompiler())
    def __floor__(self: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("idiv_s", num_params=2))
    def __floordiv__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("ige_s"))
    def __ge__(self: int, other: int) -> bool:
        ...

    @guppy.hugr_op(builtins, int_op("igt_s"))
    def __gt__(self: int, other: int) -> bool:
        ...

    @guppy.custom(builtins, NoopCompiler())
    def __int__(self: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("inot"))
    def __invert__(self: int) -> bool:
        ...

    @guppy.hugr_op(builtins, int_op("ile_s"))
    def __le__(self: int, other: int) -> bool:
        ...

    @guppy.hugr_op(builtins, int_op("ishl", num_params=2))  # TODO: RHS is unsigned
    def __lshift__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("ilt_s"))
    def __lt__(self: int, other: int) -> bool:
        ...

    @guppy.hugr_op(builtins, int_op("imod_s", num_params=2))
    def __mod__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("imul"))
    def __mul__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("ine"))
    def __ne__(self: int, other: int) -> bool:
        ...

    @guppy.hugr_op(builtins, int_op("ineg"))
    def __neg__(self: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("ior"))
    def __or__(self: int, other: int) -> int:
        ...

    @guppy.custom(builtins, NoopCompiler())
    def __pos__(self: int) -> int:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="ipow"))  # TODO
    def __pow__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("iadd"), ReversingChecker())
    def __radd__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("rand"), ReversingChecker())
    def __rand__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("idivmod_s", num_params=2), ReversingChecker())
    def __rdivmod__(self: int, other: int) -> tuple[int, int]:
        ...

    @guppy.hugr_op(builtins, int_op("idiv_s", num_params=2), ReversingChecker())
    def __rfloordiv__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(
        builtins, int_op("ishl", num_params=2), ReversingChecker()
    )  # TODO: RHS is unsigned
    def __rlshift__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("imod_s", num_params=2), ReversingChecker())
    def __rmod__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("imul"), ReversingChecker())
    def __rmul__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("ior"), ReversingChecker())
    def __ror__(self: int, other: int) -> int:
        ...

    @guppy.custom(builtins, NoopCompiler())
    def __round__(self: int) -> int:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="ipow"), ReversingChecker())  # TODO
    def __rpow__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(
        builtins, int_op("ishr", num_params=2), ReversingChecker()
    )  # TODO: RHS is unsigned
    def __rrshift__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("ishr", num_params=2))  # TODO: RHS is unsigned
    def __rshift__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("isub"), ReversingChecker())
    def __rsub__(self: int, other: int) -> int:
        ...

    @guppy.custom(builtins, IntTruedivCompiler(), ReversingChecker())
    def __rtruediv__(self: int, other: int) -> float:
        ...

    @guppy.hugr_op(builtins, int_op("ixor"), ReversingChecker())
    def __rxor__(self: int, other: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("isub"))
    def __sub__(self: int, other: int) -> int:
        ...

    @guppy.custom(builtins, IntTruedivCompiler())
    def __truediv__(self: int, other: int) -> float:
        ...

    @guppy.custom(builtins, NoopCompiler())
    def __trunc__(self: int) -> int:
        ...

    @guppy.hugr_op(builtins, int_op("ixor"))
    def __xor__(self: int, other: int) -> int:
        ...


@guppy.type(builtins, hugr_float_type, name="float", bound=tys.TypeBound.Copyable)
class Float:
    @guppy.hugr_op(builtins, float_op("fabs"), CoercingChecker())
    def __abs__(self: float) -> float:
        ...

    @guppy.hugr_op(builtins, float_op("fadd"), CoercingChecker())
    def __add__(self: float, other: float) -> float:
        ...

    @guppy.custom(builtins, FloatBoolCompiler(), CoercingChecker())
    def __bool__(self: float) -> bool:
        ...

    @guppy.hugr_op(builtins, float_op("fceil"), CoercingChecker())
    def __ceil__(self: float) -> float:
        ...

    @guppy.custom(builtins, FloatDivmodCompiler(), CoercingChecker())
    def __divmod__(self: float, other: float) -> tuple[float, float]:
        ...

    @guppy.hugr_op(builtins, float_op("feq"), CoercingChecker())
    def __eq__(self: float, other: float) -> bool:
        ...

    @guppy.custom(builtins, NoopCompiler(), CoercingChecker())
    def __float__(self: float) -> float:
        ...

    @guppy.hugr_op(builtins, float_op("ffloor"), CoercingChecker())
    def __floor__(self: float) -> float:
        ...

    @guppy.custom(builtins, FloatFloordivCompiler(), CoercingChecker())
    def __floordiv__(self: float, other: float) -> float:
        ...

    @guppy.hugr_op(builtins, float_op("fge"), CoercingChecker())
    def __ge__(self: float, other: float) -> bool:
        ...

    @guppy.hugr_op(builtins, float_op("fgt"), CoercingChecker())
    def __gt__(self: float, other: float) -> bool:
        ...

    @guppy.hugr_op(
        builtins, float_op("trunc_s", "arithmetic.conversions"), CoercingChecker()
    )
    def __int__(self: float) -> int:
        ...

    @guppy.hugr_op(builtins, float_op("fle"), CoercingChecker())
    def __le__(self: float, other: float) -> bool:
        ...

    @guppy.hugr_op(builtins, float_op("flt"), CoercingChecker())
    def __lt__(self: float, other: float) -> bool:
        ...

    @guppy.custom(builtins, FloatModCompiler(), CoercingChecker())
    def __mod__(self: float, other: float) -> float:
        ...

    @guppy.hugr_op(builtins, float_op("fmul"), CoercingChecker())
    def __mul__(self: float, other: float) -> float:
        ...

    @guppy.hugr_op(builtins, float_op("fne"), CoercingChecker())
    def __ne__(self: float, other: float) -> bool:
        ...

    @guppy.hugr_op(builtins, float_op("fneg"), CoercingChecker())
    def __neg__(self: float, other: float) -> float:
        ...

    @guppy.custom(builtins, NoopCompiler(), CoercingChecker())
    def __pos__(self: float) -> float:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="fpow"))  # TODO
    def __pow__(self: float, other: float) -> float:
        ...

    @guppy.hugr_op(builtins, float_op("fadd"), ReversingChecker(CoercingChecker()))
    def __radd__(self: float, other: float) -> float:
        ...

    @guppy.custom(builtins, FloatDivmodCompiler(), ReversingChecker(CoercingChecker()))
    def __rdivmod__(self: float, other: float) -> tuple[float, float]:
        ...

    @guppy.custom(
        builtins, FloatFloordivCompiler(), ReversingChecker(CoercingChecker())
    )
    def __rfloordiv__(self: float, other: float) -> float:
        ...

    @guppy.custom(builtins, FloatModCompiler(), ReversingChecker(CoercingChecker()))
    def __rmod__(self: float, other: float) -> float:
        ...

    @guppy.hugr_op(builtins, float_op("fmul"), ReversingChecker(CoercingChecker()))
    def __rmul__(self: float, other: float) -> float:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="fround"))  # TODO
    def __round__(self: float) -> float:
        ...

    @guppy.hugr_op(
        builtins, ops.DummyOp(name="fpow"), ReversingChecker(DefaultCallChecker())
    )  # TODO
    def __rpow__(self: float, other: float) -> float:
        ...

    @guppy.hugr_op(builtins, float_op("fsub"), ReversingChecker(CoercingChecker()))
    def __rsub__(self: float, other: float) -> float:
        ...

    @guppy.hugr_op(builtins, float_op("fdiv"), ReversingChecker(CoercingChecker()))
    def __rtruediv__(self: float, other: float) -> float:
        ...

    @guppy.hugr_op(builtins, float_op("fsub"), CoercingChecker())
    def __sub__(self: float, other: float) -> float:
        ...

    @guppy.hugr_op(builtins, float_op("fdiv"), CoercingChecker())
    def __truediv__(self: float, other: float) -> float:
        ...

    @guppy.hugr_op(
        builtins, float_op("trunc_s", "arithmetic.conversions"), CoercingChecker()
    )
    def __trunc__(self: float) -> float:
        ...


@guppy.extend_type(builtins, ListType)
class List:
    @guppy.hugr_op(builtins, ops.DummyOp(name="Concat"))
    def __add__(self: list[T], other: list[T]) -> list[T]:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="IsEmpty"))
    def __bool__(self: list[T]) -> bool:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="Contains"))
    def __contains__(self: list[T], el: T) -> bool:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="AssertEmpty"))
    def __end__(self: list[T]) -> None:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="Lookup"))
    def __getitem__(self: list[T], idx: int) -> T:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="IsNonEmpty"))
    def __hasnext__(self: list[T]) -> tuple[bool, list[T]]:
        ...

    @guppy.custom(builtins, NoopCompiler())
    def __iter__(self: list[T]) -> list[T]:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="Length"))
    def __len__(self: list[T]) -> int:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="Repeat"))
    def __mul__(self: list[T], other: int) -> list[T]:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="Pop"))
    def __next__(self: list[T]) -> tuple[T, list[T]]:
        ...

    @guppy.custom(builtins, checker=FailingChecker("Guppy lists are immutable"))
    def __setitem__(self: list[T], idx: int, value: T) -> None:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="Append"), ReversingChecker())
    def __radd__(self: list[T], other: list[T]) -> list[T]:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="Repeat"), ReversingChecker())
    def __rmul__(self: list[T], other: int) -> list[T]:
        ...

    @guppy.custom(builtins, checker=FailingChecker("Guppy lists are immutable"))
    def append(self: list[T], elt: T) -> None:
        ...

    @guppy.custom(builtins, checker=FailingChecker("Guppy lists are immutable"))
    def clear(self: list[T]) -> None:
        ...

    @guppy.custom(builtins, NoopCompiler())  # Can be noop since lists are immutable
    def copy(self: list[T]) -> list[T]:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="Count"))
    def count(self: list[T], elt: T) -> int:
        ...

    @guppy.custom(builtins, checker=FailingChecker("Guppy lists are immutable"))
    def extend(self: list[T], seq: None) -> None:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="Find"))
    def index(self: list[T], elt: T) -> int:
        ...

    @guppy.custom(builtins, checker=FailingChecker("Guppy lists are immutable"))
    def pop(self: list[T], idx: int) -> None:
        ...

    @guppy.custom(builtins, checker=FailingChecker("Guppy lists are immutable"))
    def remove(self: list[T], elt: T) -> None:
        ...

    @guppy.custom(builtins, checker=FailingChecker("Guppy lists are immutable"))
    def reverse(self: list[T]) -> None:
        ...

    @guppy.custom(builtins, checker=FailingChecker("Guppy lists are immutable"))
    def sort(self: list[T]) -> None:
        ...


linst = list


@guppy.extend_type(builtins, LinstType)
class Linst:
    @guppy.hugr_op(builtins, ops.DummyOp(name="Append"))
    def __add__(self: linst[L], other: linst[L]) -> linst[L]:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="AssertEmpty"))
    def __end__(self: linst[L]) -> None:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="IsNonempty"))
    def __hasnext__(self: linst[L]) -> tuple[bool, linst[L]]:
        ...

    @guppy.custom(builtins, NoopCompiler())
    def __iter__(self: linst[L]) -> linst[L]:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="Length"))
    def __len__(self: linst[L]) -> tuple[int, linst[L]]:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="Pop"))
    def __next__(self: linst[L]) -> tuple[L, linst[L]]:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="Append"), ReversingChecker())
    def __radd__(self: linst[L], other: linst[L]) -> linst[L]:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="Repeat"), ReversingChecker())
    def __rmul__(self: linst[L], other: int) -> linst[L]:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="Push"))
    def append(self: linst[L], elt: L) -> linst[L]:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="PopAt"))
    def pop(self: linst[L], idx: int) -> tuple[L, linst[L]]:
        ...

    @guppy.hugr_op(builtins, ops.DummyOp(name="Reverse"))
    def reverse(self: linst[L]) -> linst[L]:
        ...

    @guppy.custom(builtins, checker=FailingChecker("Guppy lists are immutable"))
    def sort(self: list[T]) -> None:
        ...


@guppy.custom(builtins, checker=DunderChecker("__abs__"), higher_order_value=False)
def abs(x):
    ...


@guppy.custom(
    builtins, name="bool", checker=DunderChecker("__bool__"), higher_order_value=False
)
def _bool(x):
    ...


@guppy.custom(builtins, checker=CallableChecker(), higher_order_value=False)
def callable(x):
    ...


@guppy.custom(
    builtins, checker=DunderChecker("__divmod__", num_args=2), higher_order_value=False
)
def divmod(x, y):
    ...


@guppy.custom(
    builtins, name="float", checker=DunderChecker("__float__"), higher_order_value=False
)
def _float(x, y):
    ...


@guppy.custom(
    builtins, name="int", checker=DunderChecker("__int__"), higher_order_value=False
)
def _int(x):
    ...


@guppy.custom(builtins, checker=DunderChecker("__len__"), higher_order_value=False)
def len(x):
    ...


@guppy.custom(
    builtins, checker=DunderChecker("__pow__", num_args=2), higher_order_value=False
)
def pow(x, y):
    ...


@guppy.custom(builtins, checker=DunderChecker("__round__"), higher_order_value=False)
def round(x):
    ...


# Python builtins that are not supported yet:


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def aiter(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def all(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def anext(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def any(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def bin(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def breakpoint(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def bytearray(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def bytes(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def chr(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def classmethod(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def compile(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def complex(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def delattr(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def dict(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def dir(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def enumerate(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def eval(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def exec(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def filter(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def format(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def forozenset(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def getattr(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def globals():
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def hasattr(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def hash(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def help(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def hex(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def id(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def input(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def isinstance(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def issubclass(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def iter(x):
    ...


@guppy.custom(
    builtins, checker=UnsupportedChecker(), name="list", higher_order_value=False
)
def _list(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def locals(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def map(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def max(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def memoryview(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def min(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def next(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def object(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def oct(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def open(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def ord(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def print(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def property(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def range(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def repr(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def reversed(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def set(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def setattr(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def slice(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def sorted(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def staticmethod(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def str(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def sum(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def super(x):
    ...


@guppy.custom(
    builtins, name="tuple", checker=UnsupportedChecker(), higher_order_value=False
)
def _tuple(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def type(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def vars(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def zip(x):
    ...


@guppy.custom(builtins, checker=UnsupportedChecker(), higher_order_value=False)
def __import__(x):
    ...



from typing import Any, Literal

from pydantic import BaseModel

from guppylang.hugr.ops import NodeID, OpType

Port = tuple[NodeID, int | None]  # (node, offset)
Edge = tuple[Port, Port]


class RawHugr(BaseModel):
    version: Literal["v0"] = "v0"
    nodes: list[OpType]
    edges: list[Edge]

    def to_json(self) -> str:
        """Return a JSON representation of the Hugr."""
        return self.model_dump_json()

    @classmethod
    def load_json(cls, json: dict[Any, Any]) -> "RawHugr":
        """Decode a JSON-encoded Hugr."""
        return cls(**json)



import ast
import functools
import sys
import textwrap
from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from typing import Any, TypeVar, cast

from guppylang.ast_util import AstNode, get_file, get_line_offset, get_source
from guppylang.gtypes import BoundTypeVar, ExistentialTypeVar, FunctionType, GuppyType
from guppylang.hugr.hugr import Node, OutPortV

# Whether the interpreter should exit when a Guppy error occurs
EXIT_ON_ERROR: bool = True


@dataclass(frozen=True)
class SourceLoc:
    """A source location associated with an AST node.

    This class translates the location data provided by the ast module into a location
    inside the file.
    """

    file: str
    line: int
    col: int
    ast_node: AstNode | None

    @staticmethod
    def from_ast(node: AstNode) -> "SourceLoc":
        file, line_offset = get_file(node), get_line_offset(node)
        assert file is not None
        assert line_offset is not None
        return SourceLoc(file, line_offset + node.lineno - 1, node.col_offset, node)

    def __str__(self) -> str:
        return f"{self.line}:{self.col}"

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, SourceLoc):
            return NotImplemented
        return (self.line, self.col) < (other.line, other.col)


@dataclass
class GuppyError(Exception):
    """General Guppy error tied to a node in the AST.

    The error message can also refer to AST locations using format placeholders `{0}`,
    `{1}`, etc. and passing the corresponding AST nodes to `locs_in_msg`."""

    raw_msg: str
    location: AstNode | None = None
    # The message can also refer to AST locations using format placeholders `{0}`, `{1}`
    locs_in_msg: Sequence[AstNode | None] = field(default_factory=list)

    def get_msg(self) -> str:
        """Returns the message associated with this error.

        A line offset is needed to translate AST locations mentioned in the message into
        source locations in the actual file."""
        return self.raw_msg.format(
            *(
                SourceLoc.from_ast(loc) if loc is not None else "???"
                for loc in self.locs_in_msg
            )
        )


class GuppyTypeError(GuppyError):
    """Special Guppy exception for type errors."""


class GuppyTypeInferenceError(GuppyError):
    """Special Guppy exception for type inference errors."""


class MissingModuleError(GuppyError):
    """Special Guppy exception for operations that require a guppy module."""


class InternalGuppyError(Exception):
    """Exception for internal problems during compilation."""


class UndefinedPort(OutPortV):
    """Dummy port for undefined variables.

    Raises an `InternalGuppyError` if one tries to access one of its properties.
    """

    def __init__(self, ty: GuppyType):
        self._ty = ty

    @property
    def ty(self) -> GuppyType:
        return self._ty

    @property
    def node(self) -> Node:
        raise InternalGuppyError("Tried to access undefined Port")

    @property
    def offset(self) -> int:
        raise InternalGuppyError("Tried to access undefined Port")


class UnknownFunctionType(FunctionType):
    """Dummy function type for custom functions without an expressible type.

    Raises an `InternalGuppyError` if one tries to access one of its members.
    """

    def __init__(self) -> None:
        pass

    @property
    def args(self) -> Sequence[GuppyType]:
        raise InternalGuppyError("Tried to access unknown function type")

    @property
    def returns(self) -> GuppyType:
        raise InternalGuppyError("Tried to access unknown function type")

    @property
    def args_names(self) -> Sequence[str] | None:
        raise InternalGuppyError("Tried to access unknown function type")

    @property
    def quantified(self) -> Sequence[BoundTypeVar]:
        raise InternalGuppyError("Tried to access unknown function type")

    @property
    def unsolved_vars(self) -> set[ExistentialTypeVar]:
        return set()


def format_source_location(
    loc: ast.AST,
    num_lines: int = 3,
    indent: int = 4,
) -> str:
    """Creates a pretty banner to show source locations for errors."""
    source, line_offset = get_source(loc), get_line_offset(loc)
    assert source is not None
    assert line_offset is not None
    source_lines = source.splitlines(keepends=True)
    end_col_offset = loc.end_col_offset
    if end_col_offset is None or (loc.end_lineno and loc.end_lineno > loc.lineno):
        end_col_offset = len(source_lines[loc.lineno - 1]) - 1
    s = "".join(source_lines[max(loc.lineno - num_lines, 0) : loc.lineno]).rstrip()
    s += "\n" + loc.col_offset * " " + (end_col_offset - loc.col_offset) * "^"
    s = textwrap.dedent(s).splitlines()
    # Add line numbers
    line_numbers = [
        str(line_offset + loc.lineno - i) + ":" for i in range(num_lines, 0, -1)
    ]
    longest = max(len(ln) for ln in line_numbers)
    prefixes = [ln + " " * (longest - len(ln) + indent) for ln in line_numbers]
    res = "".join(prefix + line + "\n" for prefix, line in zip(prefixes, s[:-1]))
    res += (longest + indent) * " " + s[-1]
    return res


FuncT = TypeVar("FuncT", bound=Callable[..., Any])


def pretty_errors(f: FuncT) -> FuncT:
    """Decorator to print custom error banners when a `GuppyError` occurs."""

    @functools.wraps(f)
    def pretty_errors_wrapped(*args: Any, **kwargs: Any) -> Any:
        try:
            return f(*args, **kwargs)
        except GuppyError as err:
            # Reraise if we're missing a location
            if not err.location:
                raise
            loc = err.location
            file, line_offset = get_file(loc), get_line_offset(loc)
            assert file is not None
            assert line_offset is not None
            line = line_offset + loc.lineno - 1
            print(  # noqa: T201
                f"Guppy compilation failed. Error in file {file}:{line}\n\n"
                f"{format_source_location(loc)}\n"
                f"{err.__class__.__name__}: {err.get_msg()}",
                file=sys.stderr,
            )
            if EXIT_ON_ERROR:
                sys.exit(1)
            return None

    return cast(FuncT, pretty_errors_wrapped)



from dataclasses import dataclass

from guppylang.ast_util import AstNode
from guppylang.checker.func_checker import CheckedFunction, DefinedFunction
from guppylang.compiler.cfg_compiler import compile_cfg
from guppylang.compiler.core import (
    CompiledFunction,
    CompiledGlobals,
    DFContainer,
    PortVariable,
)
from guppylang.gtypes import FunctionType, Inst, type_to_row
from guppylang.hugr.hugr import DFContainingVNode, Hugr, OutPortV
from guppylang.nodes import CheckedNestedFunctionDef


@dataclass
class CompiledFunctionDef(DefinedFunction, CompiledFunction):
    node: DFContainingVNode

    def load(
        self, dfg: DFContainer, graph: Hugr, globals: CompiledGlobals, node: AstNode
    ) -> OutPortV:
        return graph.add_load_constant(self.node.out_port(0), dfg.node).out_port(0)

    def compile_call(
        self,
        args: list[OutPortV],
        type_args: Inst,
        dfg: DFContainer,
        graph: Hugr,
        globals: CompiledGlobals,
        node: AstNode,
    ) -> list[OutPortV]:
        # TODO: Hugr should probably allow us to pass type args to `Call`, so we can
        #   avoid loading the function to manually add a `TypeApply`
        if type_args:
            func = graph.add_load_constant(self.node.out_port(0), dfg.node)
            func = graph.add_type_apply(func.out_port(0), type_args, dfg.node)
            call = graph.add_indirect_call(func.out_port(0), args, dfg.node)
        else:
            call = graph.add_call(self.node.out_port(0), args, dfg.node)
        return [call.out_port(i) for i in range(len(type_to_row(self.ty.returns)))]


def compile_global_func_def(
    func: CheckedFunction,
    def_node: DFContainingVNode,
    graph: Hugr,
    globals: CompiledGlobals,
) -> CompiledFunctionDef:
    """Compiles a top-level function definition to Hugr."""
    _, ports = graph.add_input_with_ports(list(func.ty.args), def_node)
    cfg_node = graph.add_cfg(def_node, ports)
    compile_cfg(func.cfg, graph, cfg_node, globals)

    # Add output node for the cfg
    graph.add_output(
        inputs=[cfg_node.add_out_port(ty) for ty in type_to_row(func.cfg.output_ty)],
        parent=def_node,
    )

    return CompiledFunctionDef(func.name, func.ty, func.defined_at, None, def_node)


def compile_local_func_def(
    func: CheckedNestedFunctionDef,
    dfg: DFContainer,
    graph: Hugr,
    globals: CompiledGlobals,
) -> PortVariable:
    """Compiles a local (nested) function definition to Hugr."""
    assert func.ty.arg_names is not None

    # Pick an order for the captured variables
    captured = list(func.captured.values())

    # Prepend captured variables to the function arguments
    closure_ty = FunctionType(
        [v.ty for v in captured] + list(func.ty.args),
        func.ty.returns,
        [v.name for v in captured] + list(func.ty.arg_names),
    )

    def_node = graph.add_def(closure_ty, dfg.node, func.name)
    def_input, input_ports = graph.add_input_with_ports(list(closure_ty.args), def_node)

    # If we have captured variables and the body contains a recursive occurrence of
    # the function itself, then we provide the partially applied function as a local
    # variable
    if len(captured) > 0 and func.name in func.cfg.live_before[func.cfg.entry_bb]:
        loaded = graph.add_load_constant(def_node.out_port(0), def_node).out_port(0)
        partial = graph.add_partial(
            loaded, [def_input.out_port(i) for i in range(len(captured))], def_node
        )
        input_ports.append(partial.out_port(0))
        func.cfg.input_tys.append(func.ty)
    else:
        # Otherwise, we treat the function like a normal global variable
        globals = globals | {
            func.name: CompiledFunctionDef(func.name, func.ty, func, None, def_node)
        }

    # Compile the CFG
    cfg_node = graph.add_cfg(def_node, inputs=input_ports)
    compile_cfg(func.cfg, graph, cfg_node, globals)

    # Add output node for the cfg
    graph.add_output(
        inputs=[cfg_node.add_out_port(ty) for ty in type_to_row(func.cfg.output_ty)],
        parent=def_node,
    )

    # Finally, load the function into the local data-flow graph
    loaded = graph.add_load_constant(def_node.out_port(0), dfg.node).out_port(0)
    if len(captured) > 0:
        loaded = graph.add_partial(
            loaded, [dfg[v.name].port for v in captured], dfg.node
        ).out_port(0)

    return PortVariable(func.name, loaded, func)



import ast
import copy
import itertools
from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from typing import Any, NamedTuple

from guppylang.ast_util import AstNode, name_nodes_in_ast
from guppylang.gtypes import (
    BoolType,
    FunctionType,
    GuppyType,
    LinstType,
    ListType,
    NoneType,
    Subst,
    SumType,
    TupleType,
)


@dataclass
class Variable:
    """Class holding data associated with a variable."""

    name: str
    ty: GuppyType
    defined_at: AstNode | None
    used: AstNode | None


@dataclass
class CallableVariable(ABC, Variable):
    """Abstract base class for global variables that can be called."""

    ty: FunctionType

    @abstractmethod
    def check_call(
        self, args: list[ast.expr], ty: GuppyType, node: AstNode, ctx: "Context"
    ) -> tuple[ast.expr, Subst]:
        """Checks the return type of a function call against a given type."""

    @abstractmethod
    def synthesize_call(
        self, args: list[ast.expr], node: AstNode, ctx: "Context"
    ) -> tuple[ast.expr, GuppyType]:
        """Synthesizes the return type of a function call."""


@dataclass
class TypeVarDecl:
    """A declared type variable."""

    name: str
    linear: bool


PyScope = dict[str, Any]


class Globals(NamedTuple):
    """Collection of names that are available on module-level.

    Separately stores names that are bound to values (i.e. module-level functions or
    constants), to types, or to instance functions belonging to types.
    """

    values: dict[str, Variable]
    types: dict[str, type[GuppyType]]
    type_vars: dict[str, TypeVarDecl]
    python_scope: PyScope

    @staticmethod
    def default() -> "Globals":
        """Generates a `Globals` instance that is populated with all core types"""
        tys: dict[str, type[GuppyType]] = {
            FunctionType.name: FunctionType,
            TupleType.name: TupleType,
            SumType.name: SumType,
            NoneType.name: NoneType,
            BoolType.name: BoolType,
            ListType.name: ListType,
            LinstType.name: LinstType,
        }
        return Globals({}, tys, {}, {})

    def get_instance_func(self, ty: GuppyType, name: str) -> CallableVariable | None:
        """Looks up an instance function with a given name for a type.

        Returns `None` if the name doesn't exist or isn't a function.
        """
        qualname = qualified_name(ty.__class__, name)
        if qualname in self.values:
            val = self.values[qualname]
            if isinstance(val, CallableVariable):
                return val
        return None

    def __or__(self, other: "Globals") -> "Globals":
        return Globals(
            self.values | other.values,
            self.types | other.types,
            self.type_vars | other.type_vars,
            self.python_scope | other.python_scope,
        )

    def __ior__(self, other: "Globals") -> "Globals":  # noqa: PYI034
        self.values.update(other.values)
        self.types.update(other.types)
        self.type_vars.update(other.type_vars)
        return self


@dataclass
class Locals:
    """Scoped mapping from names to variables"""

    vars: dict[str, Variable]
    parent_scope: "Locals | None" = None

    def __getitem__(self, item: str) -> Variable:
        if item not in self.vars and self.parent_scope:
            return self.parent_scope[item]

        return self.vars[item]

    def __setitem__(self, key: str, value: Variable) -> None:
        self.vars[key] = value

    def __iter__(self) -> Iterator[str]:
        parent_iter = iter(self.parent_scope) if self.parent_scope else iter(())
        return itertools.chain(iter(self.vars), parent_iter)

    def __contains__(self, item: str) -> bool:
        return (item in self.vars) or (
            self.parent_scope is not None and item in self.parent_scope
        )

    def __copy__(self) -> "Locals":
        # Make a copy of the var map so that mutating the copy doesn't
        # mutate our variable mapping
        return Locals(self.vars.copy(), copy.copy(self.parent_scope))

    def keys(self) -> set[str]:
        parent_keys = self.parent_scope.keys() if self.parent_scope else set()
        return parent_keys | self.vars.keys()

    def items(self) -> Iterable[tuple[str, Variable]]:
        parent_items = (
            iter(self.parent_scope.items()) if self.parent_scope else iter(())
        )
        return itertools.chain(self.vars.items(), parent_items)


class Context(NamedTuple):
    """The type checking context."""

    globals: Globals
    locals: Locals


class DummyEvalDict(PyScope):
    """A custom dict that can be passed to `eval` to give better error messages.
    This class is used to implement the `py(...)` expression. If the user tries to
    access a Guppy variable in the Python context, we give an informative error message.
    """

    ctx: Context
    node: ast.expr

    @dataclass
    class GuppyVarUsedError(BaseException):
        """Error that is raised when the user tries to access a Guppy variable."""

        var: str
        node: ast.Name | None

    def __init__(self, ctx: Context, node: ast.expr):
        super().__init__(**ctx.globals.python_scope)
        self.ctx = ctx
        self.node = node

    def _check_item(self, key: str) -> None:
        # Catch the user trying to access Guppy variables
        if key in self.ctx.locals:
            # Find the name node in the AST where the usage occurs
            n = next((n for n in name_nodes_in_ast(self.node) if n.id == key), None)
            raise self.GuppyVarUsedError(key, n)

    def __getitem__(self, key: str) -> Any:
        self._check_item(key)
        return super().__getitem__(key)

    def __delitem__(self, key: str) -> None:
        self._check_item(key)
        super().__delitem__(key)

    def __contains__(self, key: object) -> bool:
        if isinstance(key, str) and key in self.ctx.locals:
            return True
        return super().__contains__(key)


def qualified_name(ty: type[GuppyType] | str, name: str) -> str:
    """Returns a qualified name for an instance function on a type."""
    ty_name = ty if isinstance(ty, str) else ty.name
    return f"{ty_name}.{name}"



"""Guppy standard module for quantum operations."""

# mypy: disable-error-code=empty-body

from guppylang.decorator import guppy
from guppylang.hugr import ops, tys
from guppylang.hugr.tys import TypeBound
from guppylang.module import GuppyModule
from guppylang.prelude._internal import MeasureCompiler

quantum = GuppyModule("quantum")


def quantum_op(op_name: str) -> ops.OpType:
    """Utility method to create Hugr quantum ops."""
    return ops.CustomOp(extension="quantum.tket2", op_name=op_name, args=[])


@guppy.type(
    quantum,
    tys.Opaque(extension="prelude", id="qubit", args=[], bound=TypeBound.Any),
    linear=True,
)
class Qubit:
    pass


@guppy.hugr_op(quantum, quantum_op("H"))
def h(q: Qubit) -> Qubit:
    ...


@guppy.hugr_op(quantum, quantum_op("CZ"))
def cz(control: Qubit, target: Qubit) -> tuple[Qubit, Qubit]:
    ...


@guppy.hugr_op(quantum, quantum_op("CX"))
def cx(control: Qubit, target: Qubit) -> tuple[Qubit, Qubit]:
    ...


@guppy.hugr_op(quantum, quantum_op("T"))
def t(q: Qubit) -> Qubit:
    ...


@guppy.hugr_op(quantum, quantum_op("S"))
def s(q: Qubit) -> Qubit:
    ...


@guppy.hugr_op(quantum, quantum_op("X"))
def x(q: Qubit) -> Qubit:
    ...


@guppy.hugr_op(quantum, quantum_op("Y"))
def y(q: Qubit) -> Qubit:
    ...


@guppy.hugr_op(quantum, quantum_op("Z"))
def z(q: Qubit) -> Qubit:
    ...


@guppy.hugr_op(quantum, quantum_op("Tdg"))
def tdg(q: Qubit) -> Qubit:
    ...


@guppy.hugr_op(quantum, quantum_op("Sdg"))
def sdg(q: Qubit) -> Qubit:
    ...


@guppy.hugr_op(quantum, quantum_op("ZZMax"))
def zz_max(q: Qubit) -> Qubit:
    ...


@guppy.hugr_op(quantum, quantum_op("Measure"))
def measure_return(q: Qubit) -> tuple[Qubit, bool]:
    ...


@guppy.hugr_op(quantum, quantum_op("RzF64"))
def rz(q: Qubit, angle: float) -> Qubit:
    ...


@guppy.hugr_op(quantum, quantum_op("RxF64"))
def rx(q: Qubit, angle: float) -> Qubit:
    ...


@guppy.hugr_op(quantum, quantum_op("RzF64"))
def phased_x(q: Qubit, angle1: float, angle2: float) -> Qubit:
    ...


@guppy.hugr_op(quantum, quantum_op("RzF64"))
def zz_phase(q1: Qubit, q2: Qubit, angle: float) -> tuple[Qubit, Qubit]:
    ...


@guppy.hugr_op(quantum, quantum_op("RzF64"))
def tk1(q: Qubit, angle1: float, angle2: float, angle3: float) -> Qubit:
    ...


@guppy.hugr_op(quantum, quantum_op("QAlloc"), name="Qubit")
def _Qubit() -> Qubit:
    ...


@guppy.hugr_op(quantum, quantum_op("QFree"))
def discard(q: Qubit) -> None:
    ...


@guppy.hugr_op(quantum, quantum_op("Reset"))
def reset(q: Qubit) -> Qubit:
    ...


@guppy.custom(quantum, MeasureCompiler())
def measure(q: Qubit) -> bool:
    ...



import functools
import inspect
from collections.abc import Callable, Iterator, Sequence
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Any, ClassVar, TypeVar

from guppylang.ast_util import AstNode, has_empty_body
from guppylang.custom import (
    CustomCallChecker,
    CustomCallCompiler,
    CustomFunction,
    DefaultCallChecker,
    DefaultCallCompiler,
    OpCompiler,
)
from guppylang.error import GuppyError, MissingModuleError, pretty_errors
from guppylang.gtypes import GuppyType, TypeTransformer
from guppylang.hugr import ops, tys
from guppylang.hugr.hugr import Hugr
from guppylang.module import GuppyModule, PyFunc, parse_py_func

FuncDecorator = Callable[[PyFunc], PyFunc | Hugr]
CustomFuncDecorator = Callable[[PyFunc], CustomFunction]
ClassDecorator = Callable[[type], type]


@dataclass(frozen=True)
class ModuleIdentifier:
    """Identifier for the Python file/module that called the decorator."""

    filename: Path
    module: ModuleType | None

    @property
    def name(self) -> str:
        """Returns a user-friendly name for the caller.

        If the called is not a function, uses the file name.
        """
        if self.module is not None:
            return str(self.module.__name__)
        return self.filename.name


class _Guppy:
    """Class for the `@guppy` decorator."""

    # The currently-alive GuppyModules, associated with a Python file/module.
    #
    # Only contains **uncompiled** modules.
    _modules: dict[ModuleIdentifier, GuppyModule]

    def __init__(self) -> None:
        self._modules = {}

    @pretty_errors
    def __call__(self, arg: PyFunc | GuppyModule) -> FuncDecorator:
        """Decorator to annotate Python functions as Guppy code.

        Optionally, the `GuppyModule` in which the function should be placed can
        be passed to the decorator.
        """

        def make_dummy(wraps: PyFunc) -> Callable[..., Any]:
            @functools.wraps(wraps)
            def dummy(*args: Any, **kwargs: Any) -> Any:
                raise GuppyError(
                    "Guppy functions can only be called in a Guppy context"
                )

            return dummy

        if not isinstance(arg, GuppyModule):
            # Decorator used without any arguments.
            # We default to a module associated with the caller of the decorator.
            f = arg

            caller = self._get_python_caller(f)
            if caller not in self._modules:
                self._modules[caller] = GuppyModule(caller.name)
            module = self._modules[caller]
            module.register_func_def(f)
            return make_dummy(f)

        if isinstance(arg, GuppyModule):
            # Module passed.
            def dec(f: Callable[..., Any]) -> Callable[..., Any]:
                arg.register_func_def(f)
                return make_dummy(f)

            return dec

        raise ValueError(f"Invalid arguments to `@guppy` decorator: {arg}")

    def _get_python_caller(self, fn: PyFunc | None = None) -> ModuleIdentifier:
        """Returns an identifier for the Python file/module that called the decorator.

        :param fn: Optional. The function that was decorated.
        """
        if fn is not None:
            filename = inspect.getfile(fn)
            module = inspect.getmodule(fn)
        else:
            for s in inspect.stack():
                if s.filename != __file__:
                    filename = s.filename
                    module = inspect.getmodule(s.frame)
                    break
            else:
                raise GuppyError("Could not find a caller for the `@guppy` decorator")
        return ModuleIdentifier(Path(filename), module)

    @pretty_errors
    def extend_type(self, module: GuppyModule, ty: type[GuppyType]) -> ClassDecorator:
        """Decorator to add new instance functions to a type."""
        module._instance_func_buffer = {}

        def dec(c: type) -> type:
            module._register_buffered_instance_funcs(ty)
            return c

        return dec

    @pretty_errors
    def type(
        self,
        module: GuppyModule,
        hugr_ty: tys.Type,
        name: str = "",
        linear: bool = False,
        bound: tys.TypeBound | None = None,
    ) -> ClassDecorator:
        """Decorator to annotate a class definitions as Guppy types.

        Requires the static Hugr translation of the type. Additionally, the type can be
        marked as linear. All `@guppy` annotated functions on the class are turned into
        instance functions.
        """
        module._instance_func_buffer = {}

        def dec(c: type) -> type:
            _name = name or c.__name__

            @dataclass(frozen=True)
            class NewType(GuppyType):
                args: Sequence[GuppyType]
                name: ClassVar[str] = _name

                @staticmethod
                def build(*args: GuppyType, node: AstNode | None = None) -> "GuppyType":
                    # At the moment, custom types don't support type arguments.
                    if len(args) > 0:
                        raise GuppyError(
                            f"Type `{_name}` does not accept type parameters.", node
                        )
                    return NewType([])

                @property
                def type_args(self) -> Iterator[GuppyType]:
                    return iter(self.args)

                @property
                def linear(self) -> bool:
                    return linear

                def to_hugr(self) -> tys.Type:
                    return hugr_ty

                def hugr_bound(self) -> tys.TypeBound:
                    return bound or super().hugr_bound()

                def transform(self, transformer: TypeTransformer) -> GuppyType:
                    return transformer.transform(self) or NewType(
                        [ty.transform(transformer) for ty in self.args]
                    )

                def __str__(self) -> str:
                    return _name

            NewType.__name__ = name
            NewType.__qualname__ = _name
            module.register_type(_name, NewType)
            module._register_buffered_instance_funcs(NewType)
            c._guppy_type = NewType  # type: ignore[attr-defined]
            return c

        return dec

    @pretty_errors
    def type_var(self, module: GuppyModule, name: str, linear: bool = False) -> TypeVar:
        """Creates a new type variable in a module."""
        module.register_type_var(name, linear)
        return TypeVar(name)

    @pretty_errors
    def custom(
        self,
        module: GuppyModule,
        compiler: CustomCallCompiler | None = None,
        checker: CustomCallChecker | None = None,
        higher_order_value: bool = True,
        name: str = "",
    ) -> CustomFuncDecorator:
        """Decorator to add custom typing or compilation behaviour to function decls.

        Optionally, usage of the function as a higher-order value can be disabled. In
        that case, the function signature can be omitted if a custom call compiler is
        provided.
        """

        def dec(f: PyFunc) -> CustomFunction:
            func_ast = parse_py_func(f)
            if not has_empty_body(func_ast):
                raise GuppyError(
                    "Body of custom function declaration must be empty",
                    func_ast.body[0],
                )
            call_checker = checker or DefaultCallChecker()
            func = CustomFunction(
                name or func_ast.name,
                func_ast,
                compiler or DefaultCallCompiler(),
                call_checker,
                higher_order_value,
            )
            call_checker.func = func
            module.register_custom_func(func)
            return func

        return dec

    def hugr_op(
        self,
        module: GuppyModule,
        op: ops.OpType,
        checker: CustomCallChecker | None = None,
        higher_order_value: bool = True,
        name: str = "",
    ) -> CustomFuncDecorator:
        """Decorator to annotate function declarations as HUGR ops."""
        return self.custom(module, OpCompiler(op), checker, higher_order_value, name)

    def declare(self, module: GuppyModule) -> FuncDecorator:
        """Decorator to declare functions"""

        def dec(f: Callable[..., Any]) -> Callable[..., Any]:
            module.register_func_decl(f)

            @functools.wraps(f)
            def dummy(*args: Any, **kwargs: Any) -> Any:
                raise GuppyError(
                    "Guppy functions can only be called in a Guppy context"
                )

            return dummy

        return dec

    def load(self, m: ModuleType | GuppyModule) -> None:
        caller = self._get_python_caller()
        if caller not in self._modules:
            self._modules[caller] = GuppyModule(caller.name)
        module = self._modules[caller]
        module.load(m)

    def take_module(self, id: ModuleIdentifier | None = None) -> GuppyModule:
        """Returns the local GuppyModule, removing it from the local state."""
        orig_id = id
        if id is None:
            id = self._get_python_caller()
        if id not in self._modules:
            err = (
                f"Module {orig_id.name} not found."
                if orig_id
                else "No Guppy functions or types defined in this module."
            )
            raise MissingModuleError(err)
        return self._modules.pop(id)

    def compile_module(self, id: ModuleIdentifier | None = None) -> Hugr | None:
        """Compiles the local module into a Hugr."""
        module = self.take_module(id)
        if not module:
            err = (
                f"Module {id.name} not found."
                if id
                else "No Guppy functions or types defined in this module."
            )
            raise MissingModuleError(err)
        return module.compile()

    def registered_modules(self) -> list[ModuleIdentifier]:
        """Returns a list of all currently registered modules for local contexts."""
        return list(self._modules.keys())


guppy = _Guppy()
